/*
Extend HTML Document
*/

(function () {
'use strict'


ff.apply(ff, {

	isTouch: 'ontouchstart' in window,

	IE: Number(navigator.userAgent.select(/MSIE (\d+)/, 1)),

	ready: function (handler, scope) {
		if (document.readyState === 'complete') {
			handler.call(scope)
		}
		else {
			document.on('DOMContentLoaded', handler, scope)
		}

		return ff
	}
})


if (!ff.IE) {
	ff.IE = Number(navigator.userAgent.select(/Trident\/(\d+)/, 1) || NaN) + 4
}



var map = {
	_default: [0, '', ''],
	legend: [1, '<fieldset>', '</fieldset>'],
	tr: [2, '<table><tbody>', '</tbody></table>'],
	col: [2, '<table><tbody></tbody><colgroup>', '</colgroup></table>']
}

map.td =
	map.th = [3, '<table><tbody><tr>', '</tr></tbody></table>']

map.option =
	map.optgroup = [1, '<select multiple="multiple">', '</select>']

map.thead =
	map.tbody =
	map.colgroup =
	map.caption =
	map.tfoot = [1, '<table>', '</table>']

map.g =
	map.defs =
	map.symbol =
	map.use =
	map.image =
	map.text =
	map.circle =
	map.ellipse =
	map.line =
	map.path =
	map.polygon =
	map.polyline =
	map.rect = [
1,
'<svg '
+ 'xmlns="http://www.w3.org/2000/svg" '
+ 'xmlns:xlink="http://www.w3.org/1999/xlink" '
+ 'xmlns:ev="http://www.w3.org/2001/xml-events"'
+ 'version="1.1">',
'</svg>'
]



ff.define(Document.prototype, {

	TAG_RE: /<([\w:]+)/,

	WRAP_MAP: map,

	//why not use insertAdjacentHTML? sometimes we want to bind some event after create element
	//if use insertAdjacentHTML, dom will be forced to reflow after do some queries
	//can't used to load js files
	createFragment: function (htmlString) {
		var frag = document.createDocumentFragment(),
			tagMatch = this.TAG_RE.exec(htmlString)

		if (!tagMatch) {
			frag.appendChild(
				document.createTextNode(htmlString)
			)
		}
		else {
			var tag = tagMatch[1],
				wrap = this.WRAP_MAP[tag] || this.WRAP_MAP._default,
				depth = wrap[0],
				prefix = wrap[1],
				suffix = wrap[2],
				node = document.createElement('div')

			node.innerHTML = prefix + htmlString.trim() + suffix

			while (depth--) {
				node = node.firstChild
			}

			var child

			while (child = node.firstChild) {
				frag.appendChild(child)
			}
		}

		return frag
	},


	//can only used to create one element
	create: function (htmlString) {
		var frag = document.createFragment(htmlString)

		var child = frag.firstChild
		while (child && child.nodeType !== 1) {
			child = child.nextSibling
		}

		return child
	},


	__scrollBarWidth: undefined,


	'get scrollBarWidth': function () {
		if (this.__scrollBarWidth >= 0) {
			return this.__scrollBarWidth
		}

		var div = document.create('<div style="width:100px; height:100px; overflow:scroll; position:absolute; left:-100px; top:-100px;"></div>')
		document.body.append(div)

		this.__scrollBarWidth = div.offsetWidth - div.clientWidth

		div.destroy()

		return this.__scrollBarWidth
	}
})



ff.define(Element.prototype, {

	select: function (xpath) {
		return (this.ownerDocument || this).evaluate(xpath, this, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue
	},


	selectAll: function (xpath) {
		var items = (this.ownerDocument || this).evaluate(xpath, this, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null),
			nodes = []

		for (var i = 0; i < items.snapshotLength; i++) {
			nodes[i] = items.snapshotItem(i)
		}

		return nodes
	},


	selectValue: function (xpath) {
		if (!xpath) {
			return ''
		}

		var node = this.select(xpath)

		return node ? node.textContent : ''
	},


	query: function (sel) {
		return this.querySelector(sel)
	},


	queryAll: function (sel) {
		return ff.toArray(this.querySelectorAll(sel))
	},


	queryValue: function (sel) {
		if (!sel) {
			return ''
		}

		var node = this.querySelector(sel)

		return node ? node.textContent : ''
	},


	attr: function (name, value) {
		if (value === undefined) {
			return this.getAttribute(name)
		}
		else {
			this.setAttribute(name, value)
		}
	},


	removeAttr: function (name) {
		this.removeAttribute(name)
	},


	append: function (el) {
		if (typeof el === 'string') {
			el = this.ownerDocument.create(el)
		}

		this.appendChild(el)

		return this
	},


	prepend: function (el) {
		if (typeof el === 'string') {
			el = this.ownerDocument.create(el)
		}

		this.insertBefore(el, this.firstChild)

		return this
	},


	//insert argument el before this
	before: function (el) {
		if (!this.parentNode) {
			return null
		}

		if (typeof el === 'string') {
			el = this.ownerDocument.create(el)
		}

		this.parentNode.insertBefore(el, this)

		return this
	},


	after: function (el) {
		if (!this.parentNode) {
			return null
		}

		if (typeof el === 'string') {
			el = this.ownerDocument.create(el)
		}

		this.parentNode.insertBefore(el, this.nextElementSibling)

		return this
	},


	empty: function () {
		while (this.firstChild) {
			this.firstChild.remove()
		}
	},


	//contains width, height, left, top, right, bottom
	//if forCSS is true, delete right and bottom
	//getBoundingClientRect returns is not editable
	box: function (forBoxSetting) {
		var box = ff.clone(this.getBoundingClientRect())

		if (forBoxSetting) {
			delete box.right
			delete box.bottom
		}

		return box
	},


	__propertyMapping: (function (style) {
		var prefix = 'msTransform' in style ? 'ms' : 'mozTransform' in style ? 'Moz' : 'webkitTransform' in style ? 'webkit' : 'o'

		var o = {},
			preLen = prefix.length

		if (prefix) {
			for (var p in style) {
				if (p.startsWith(prefix)) {
					var fixedP = p[preLen].toLowerCase() + p.slice(preLen + 1)
					o[fixedP] = p
				}
			}
		}

		return o
	})(document.documentElement.style),


	//font-size -> fontSize
	//transform -> mozTransform | webkitTransform | msTransform
	fixProperty: function (p) {
		var cameledP = p.camelize(),
			mappedP = this.__propertyMapping[cameledP] || cameledP

		return mappedP
	},


	addUnit: function (value, property) {
		if (typeof value === 'number') {
			if (/(?:width|height|left|right|top|bottom|size)$/i.test(property)) {
				value += 'px'
			}
			else if (/(?:duration|delay)$/i.test(property)) {
				value += 's'
			}
			else {
				value = String(value)
			}
		}
		else {
			value = String(value)
		}

		return value
	},


	removeUnit: function (value, property) {
		if (/(?:width|height|left|right|top|bottom|size)$/i.test(property)) {
			value = Number(value.replace('px', ''))
		}
		else if (/(?:duration|delay)$/i.test(property)) {
			value = Number(value.replace('s', ''))
		}

		return value
	},


	//if value is not exist, return css property
	//if value exist, set style, auto transform number to px
	//if property is array, get all these styles
	//if property is object, set all these styles
	css: function (property, value) {
		if (typeof property === 'object') {
			var rules = property

			for (var property in rules) {
				var fixedProperty = this.fixProperty(property),
					unittedValue = this.addUnit(rules[property], property)

				this.style[fixedProperty] = unittedValue
			}

			return this
		}

		else {
			if (value === undefined) {
				var fixedProperty = this.fixProperty(property)
				value = getComputedStyle(this)[fixedProperty] || ''
				var unittlessValue = this.removeUnit(value, property)

				return unittlessValue
			}
			else {
				var fixedProperty = this.fixProperty(property),
					unittedValue = this.addUnit(value, property)

				this.style[fixedProperty] = unittedValue

				return this
			}
		}
	},


	//anchor is two chars, [tbc][lrc], default is tl, get the position for the anchor
	//if is relative, do not add box.left and box.top
	anchor: function (anchor, relative, box) {
		anchor = anchor || "tl"

		if (anchor.length === 1) {
			anchor += 'c'
		}

		box = box || this.box()

		var y = anchor[0] === 't' ? 0 : anchor[0] === 'b' ? box.height : box.height / 2,
			x = anchor[1] === 'l' ? 0 : anchor[1] === 'r' ? box.width : box.width / 2

		if (!relative) {
			x += box.left
			y += box.top
		}

		return [x, y]
	},


	//top [right] [bottom [left] -> array
	parseOffsets: function (off) {
		if (typeof off === 'string') {
			off = off.split(/\s+/)
		}
		else if (typeof off === 'number') {
			off = [off]
		}

		off = off || []
		off[0] = Number(off[0]) || 0
		off[1] = Number(off[1]) || 0
		off[2] = off[2] !== undefined ? Number(off[2]) || 0 : off[0]
		off[3] = off[3] !== undefined ? Number(off[3]) || 0 : off[1]

		return off
	},


	//align to another element by position, and offset, position of this should be fixed, please make sure this is visible
	//pos like tl-bl, bl(bl-bl), b(bc-bc) or cc-cc, or c, optional, if omit, take tl-bl, align this top left to the bottom left position of el, and auto adjust
	//off is [x, y] for mouse event, or [x, y, x, y] for to extend el, optional, if adjust position, off will be adjusted to keep the distance
	//overflowDirections will auto make this overflow to visible and not cover el
	//default pos is tl-bl, align this top left to the bottom left position of el, and auto adjust
	//el can be a MouseEvent, then pos will be tl or others, no second part
	//auto adjustment to make sure this is visible
	//return an object
	alignTo: ff.flex('object [string] [object|string|number] [string]', function (el, pos, off, overflowDirections) {
		//if set off but captured by pos, swap them
		if (!off && /\d/.test(pos)) {
			off = pos
			pos = ''
		}

		off = this.parseOffsets(off)

		//handle overflow, if have set, restore overflow
		var canOverflowX = overflowDirections.contains('x'),
			canOverflowY = overflowDirections.contains('y')

		if (canOverflowX) {
			this.css({
				width: 'auto'
			})
		}

		if (canOverflowY) {
			this.css({
				height: 'auto'
			})
		}

		var box = this.box(),
			w = box.width,
			h = box.height,
			x,
			y,
			dw = document.documentElement.clientWidth,
			dh = document.documentElement.clientHeight

		if (el instanceof MouseEvent) {
			var a = this.anchor(pos, true)

			x = el.clientX + a[0] + off[0]
			y = el.clientY + a[1] + off[1]

			x = Math.max(0, Math.min(dw - w, x))

			if (y < 0) {
				y = elBox.bottom
			}
			else if (y > dh - h) {
				y = elBox.top - h
			}
		}
		else {
			var elBox = el.box()

			//we need to set html, body{height: 100%}
			if (el === document.body || el === document.documentElement) {
				elBox.left = elBox.top = 0
				elBox.right = elBox.width
				elBox.bottom = elBox.height
			}

			elBox.left   -= off[3]
			elBox.top    -= off[0]
			elBox.right  += off[1]
			elBox.bottom += off[2]
			elBox.width  += off[1] + off[3]
			elBox.height += off[0] + off[2]

			if (!pos) {
				pos = 'tl-bl'
			}
			else if (!pos.contains('-')) {
				pos += '-' + pos
			}

			pos = pos.select(/^([a-z]+)-([a-z]+)$/)

			var p1 = pos[1],
				p2 = pos[2],
				a1 = this.anchor(p1, true, box),
				a2 = el.anchor(p2, false, elBox)

			x = a2[0] - a1[0]
			y = a2[1] - a1[1]

			var startX = x,
				startY = y

			var canSwapY = p1[0] === 't' && p2[0] === 'b' || p1[0] === 'b' && p2[0] === 't',
				canSwapX = p1[1] === 'l' && p2[1] === 'r' || p1[1] === 'r' && p2[1] === 'l',
				swapX = false,
				swapY = false,
				leftSpace = elBox.left,
				rightSpace = dw - elBox.right,
				topSpace = elBox.top,
				bottomSpace = dh - elBox.bottom,
				alignTop = p1[0] === 'b',
				alignLeft = p1[1] === 'l'

			if (x < 0 && canSwapX && leftSpace < rightSpace) {
				x = elBox.right
				swapX = true
			}
			else if (x > dw - w && canSwapX && leftSpace > rightSpace) {
				x = elBox.left - w
				swapX = true
			}

			if (y < 0 && canSwapY && topSpace < bottomSpace) {
				y = elBox.bottom
				swapY = true
			}
			else if (y + h > dh && topSpace > bottomSpace) {
				y = elBox.top - h
				swapY = true
			}

			if (swapX) {
				alignLeft = !alignLeft
			}

			if (swapY) {
				alignTop = !alignTop
			}

			if (canOverflowX && x < 0) {
				this.css('width', w = leftSpace)
			}
			else if (canOverflowX && x > dw - w) {
				this.css('width', w = rightSpace)
			}
			else if (canOverflowX) {
				var maxW = this.css('max-width')
				if (maxW > 0) {
					this.css('width', w = Math.min(w, maxW))
				}
			}

			if (canOverflowY && y < 0) {
				this.css('height', h = topSpace)
			}
			else if (canOverflowY && y > dh - h) {
				this.css('height', h = bottomSpace)
			}
			else if (canOverflowY) {
				var maxH = this.css('max-height')
				if (maxH > 0) {
					this.css('height', h = Math.min(h, maxH))
				}
			}

			x = Math.max(0, Math.min(x, dw - w))
			y = Math.max(0, Math.min(y, dh - h))
		}

		this.css({
			left: x,
			top:  y
		})

		var trangleEl = this.query('.triangle')
		if (trangleEl) {
			if (w >= elBox.width) {
				var trangleLeft = elBox.left + elBox.width / 2 - x - trangleEl.offsetWidth / 2
			}
			else {
				var trangleLeft = w / 2 - trangleEl.offsetWidth / 2
			}

			trangleEl.css('left', trangleLeft)
				
			trangleEl.removeClass('below', 'above')

			if (alignTop) {
				trangleEl.addClass('below')
			}
			else{
				trangleEl.addClass('above')
			}
		}
	}),


	__supportClassList: !!document.documentElement.classList,


	//support multy arguments
	addClass: function (cls) {
		var a = cls instanceof Array ? cls : ff.toArray(arguments)

		if (a.length > 0) {
			var classList = this.__supportClassList ? this.classList : this.className.split(/\s+/)

			classList.add.apply(classList, a)

			if (!this.__supportClassList) {
				this.className = classList.join(' ')
			}
		}

		return this
	},


	removeClass: function (cls) {
		var a = cls instanceof Array ? cls : ff.toArray(arguments)

		if (a.length > 0) {
			var classList = this.__supportClassList ? this.classList : this.className.split(/\s+/)

			classList.remove.apply(classList, a)

			if (!this.__supportClassList) {
				this.className = classList.join(' ')
			}
		}

		return this
	},


	hasClass: function (cls) {
		var a = cls instanceof Array ? cls : ff.toArray(arguments)

		if (a.length > 0) {
			var classList = this.__supportClassList ? this.classList : this.className.split(/\s+/)

			for (var i = 0; i < a.length; i++) {
				if (!classList.contains(a[i])) {
					return false
				}
			}
		}

		return true
	},


	SEL_RE: /\w+|\*|[.#][\w-]+|\[.+?\]|:[\w+-]+|>|,/g,


	ATTR_RE: /\[([\w-]+)(?:([~^$*|]?=)(['"]?)(.*?)\3)?\]/,


	__selectorCache: {},


	__selectorFns: {
		'tag': function (tag, el) {
			return el.localName === tag
		},

		'*': function (el) {
			return el.nodeType === 1
		},

		'#': function (id, el) {
			return el.id === id
		},

		'.': function (cls, el) {
			return el.hasClass(cls)
		},

		'[': {
			'': function (attr, value, el) {
				return el.hasAttribute(attr)
			},

			'=': function (attr, value, el) {
				return el.getAttribute(attr) === value
			},

			'~=': function (attr, value, el) {
				return el.getAttribute(attr).split(/\s+/g).contains(value)
			},

			'^=': function (attr, value, el) {
				return el.getAttribute(attr).startsWith(value)
			},

			'$=': function (attr, value, el) {
				return el.getAttribute(attr).endsWith(value)
			},

			'*=': function (attr, value, el) {
				return el.getAttribute(attr).contains(value)
			},

			'|=': function (attr, value, el) {
				var v = el.getAttribute(attr)
				return v === value || v.startsWith(value + '-')
			}
		},

		':': {
			'first': function (el) {
				return el.parentNode && el.parentNode.firstElementChild === el
			},

			'last': function (el) {
				return el.parentNode && el.parentNode.lastElementChild === el
			},

			'nth': function (mod, num, el) {
				var index = el.nodeIndex() + 1
				return index >= 1 && (index - num) % mod === 0
			},

			'lnth': function (mod, num, el) {
				var count = el.parentNode ? el.parentNode.children.length : -1,
					index = count - el.nodeIndex()
				return index >= 1 && (index - num) % mod === 0
			},

			'checked': function (el) {
				return el.value === true
			},

			'disabled': function (el) {
				return el.disabled === true
			},

			'hidden': function (el) {
				return el.hidden === true
			},

			'empty': function (el) {
				return !el.firstElementChild
			},

			'only': function (el) {
				return el.parentNode && el.parentNode.children.length === 1
			}
		}
	},


	//compiled: a.b > b.c, c > d -> [[[a,], b], [[c], [d]]
	__compileSelector: function (selString) {
		var sels = selString.select(this.SEL_RE, 0),
			matches = [], //must match all
			childMatches = [], //joined by >
			compiled = [] //joined by ,

		sels.push(',')

		for (var i = 0; i < sels.length; i++) {
			var sel = sels[i],
				command = sel[0]

			switch (command) {
				case '.':
				case '#':
					var value = sel.slice(1)
					matches.push(
						this.__selectorFns[command].bind(null, value)
					)
					break

				case '[':
					var atts = sel.select(this.ATTR_RE),
						att = atts[1],
						equal = atts[2] || '',
						value = atts[4] || ''

					if (!this.__selectorFns[command][equal]) {
						throw 'unknown attribute ' + sel
					}

					matches.push(
						this.__selectorFns[command][equal].bind(null, att, value)
					)
					break

				case ':':
					var presudo = sel.slice(1).select(/^[a-z]+/, 0),
						value = sel.slice(1).slice(presudo.length)

					if (presudo === 'nth' || presudo === 'lnth') {
						var mod = 1,
							num = 0

						if (!value) {
							throw 'unknown preudo ' + sel
						}
						mod = value.select(/(\d+)[+-]/, 1)
						num = value.slice(mod.length)
						mod = Number(mod) || 1
						num = Number(num) || 0

						matches.push(
							this.__selectorFns[command][presudo].bind(null, mod, num)
						)
					}

					else {
						if (value || !this.__selectorFns[command][presudo]) {
							throw 'unknown preudo ' + sel
						}

						matches.push(this.__selectorFns[command][presudo])
					}
					break

				case '*':
					matches.push(this.__selectorFns[command])
					break

				case '>':
					if (matches.length === 0) {
						matches.push(this.__selectorFns['*'])
					}
					childMatches.push(matches)
					matches = []
					break

				case ',':
					if (matches.length === 0) {
						matches.push(this.__selectorFns['*'])
					}
					childMatches.push(matches)
					compiled.push(childMatches)
					matches = []
					childMatches = []
					break

				default:
					if (!/^(?:\w+)$/.test(sel)) {
						throw 'unknown selector ' + sel
					}
					matches.push(
						this.__selectorFns['tag'].bind(null, sel)
					)
			}
		}

		this.__selectorCache[selString] = compiled

		return compiled
	},


	__isMatch: function (matches) {
		for (var i = 0; i < matches.length; i++) {
			if (!matches[i](this)) {
				return false
			}
		}

		return true
	},


	__isChildMatch: function (childMatches) {
		var node = this

		for (var i = childMatches.length - 1; i >= 0; i--) {
			var matches = childMatches[i]

			if (!node || !node.__isMatch(matches)) {
				return false
			}

			node = node.parentNode
		}

		return true
	},


	//sel must be single selector, not support like div div, the spaces between words will be removed
	//support:
	//	,  > # . div *
	//	[a], [a=b], [a~=b], [a^=b], [a$=b], a[*=b], a[|=b]
	//	:first, :last, :nth1, :nth2, :nth3+2, :lnth1(last child) :checked, :disabled, :hidden, :empty, :only
	is: function (selString) {
		var compiled = this.__selectorCache[selString] || this.__compileSelector(selString)

		for (var i = 0; i < compiled.length; i++) {
			var childMatches = compiled[i]

			if (this.__isChildMatch(childMatches)) {
				return true
			}
		}

		return false
	},


	//get one el from self and parent to match selector.
	//selector is argument for is()
	up: function (selector, until) {
		var el = this

		if (typeof until === 'string') {
			until = el.up(until)
		}

		while (el && el !== document.documentElement && el !== until) {
			if (el.is(selector)) {
				return el
			}
			el = el.parentNode
		}

		return null
	},


	//selector can be empty, return parent, otherwise return parentNode.up
	parent: function (selector) {
		if (!this.parentNode) {
			return null
		}
		else if (!selector) {
			return this.parentNode
		}
		else {
			return this.parentNode.up(selector)
		}
	},


	//first matched child
	first: function (selector) {
		if (!selector) {
			return this.firstElementChild
		}
		else {
			for (var i = 0; i < this.children.length; i++) {
				if (this.children[i].is(selector)) {
					return this.children[i]
				}
			}

			return null
		}
	},


	//last matched child
	last: function (selector) {
		if (!selector) {
			return this.lastElementChild
		}
		else {
			for (var i = this.children.length - 1; i >= 0; i--) {
				if (this.children[i].is(selector)) {
					return this.children[i]
				}
			}

			return null
		}
	},


	//gaps is number or an array
	//direction can contains x, y, or both
	scrollToView: function (direction, gaps) {
		var newScrollX,
			newScrollY,
			gapX = gaps instanceof Array ? gaps[0] : Number(gaps) || 0,
			gapY = gaps instanceof Array ? gaps[1] : Number(gaps) || 0

		var scrollParent = this.parentNode

		while (scrollParent &&
			(scrollParent.scrollWidth <= scrollParent.clientWidth) &&
			(scrollParent.scrollHeight <= scrollParent.clientHeight)) {
			scrollParent = scrollParent.parentNode
		}

		if (scrollParent === document) {
			return
		}

		if (direction.contains('x')) {
			//x is the x position after scrolling
			var x = this.offsetLeft,
				scrollX = scrollParent.scrollLeft

			//if offsetLeft is not calced by scrollParent, minus the different
			if (this.offsetParent !== scrollParent) {
				x -= scrollParent.offsetLeft
			}

			//make sure x - gapX > scrollX and x + w + gapX < scrollX + pW
			if (x - gapX < scrollX) {
				newScrollX = Math.max(0, x - gapX)
			}
			else if (x + this.offsetWidth + gapX > scrollX + scrollParent.offsetWidth) {
				newScrollX = Math.min(scrollParent.scrollWidth, x + this.offsetWidth + gapX - scrollParent.offsetWidth)
			}

			if (newScrollX === scrollX) {
				newScrollX = undefined
			}
		}

		if (direction.contains('y')) {
			var y = this.offsetTop,
				scrollY = scrollParent.scrollTop

			if (this.offsetParent !== scrollParent) {
				y -= scrollParent.offsetTop
			}

			//make sure y - gapY > scrollY and y + h + gapY < scrollY + pH
			if (y - gapY < scrollY) {
				newScrollY = Math.max(0, y - gapY)
			}
			else if (y + this.offsetHeight + gapY > scrollY + scrollParent.offsetHeight) {
				newScrollY = Math.min(scrollParent.scrollHeight, y + this.offsetHeight + gapY - scrollParent.offsetHeight)
			}

			if (newScrollY === scrollY) {
				newScrollY = undefined
			}
		}

		if (newScrollX >= 0) {
			scrollParent.scrollLeft = newScrollX
		}
		if (newScrollY >= 0) {
			scrollParent.scrollTop = newScrollY
		}
	},


	scrollToEnd: function () {
		this.scrollTop = this.scrollHeight - this.clientHeight
	},


	isScrolling: function () {
		return this.scrollHeight > this.clientHeight
	},


	'get index': function () {
		if (this.parentNode) {
			var childNodes = this.parentNode.childNodes

			for (var i = 0; i < childNodes.length; i++) {
				if (childNodes[i] === this) {
					return i
				}
			}
		}

		return -1
	},


	'get elementIndex': function () {
		if (this.parentNode) {
			var children = this.parentNode.children

			for (var i = 0; i < children.length; i++) {
				if (children[i] === this) {
					return i
				}
			}
		}

		return -1
	},


	'get outerWidth': function () {
		var w = this.offsetWidth

		if (w) {
			return w + this.css('margin-left') + this.css('margin-right')
		}
		else {
			return 0
		}
	},


	'get outerHeight': function () {
		var h = this.offsetHeight

		if (h) {
			return h + this.css('margin-top') + this.css('margin-bottom')
		}
		else {
			return 0
		}
	},


	//for elements with left and top coordinates
	//draggable conflict with html5 attribute
	makeDragable: function (follow, directions) {
		if (typeof follow === 'string') {
			follow = this.up(follow)
		}

		follow = follow || this
		directions = directions || 'xy'

		var xDragable = directions.contains('x'),
			yDragable = directions.contains('y')

		this.on('mousedown', function (e) {
			var left = follow.css('left') || 0,
				top  = follow.css('top') || 0,
				startX = e.pageX,
				startY = e.pageY,
				dw = document.documentElement.clientWidth,
				dh = document.documentElement.clientHeight,
				w  = follow.offsetWidth,
				h  = follow.offsetHeight

			var onmove = function (e) {
				var x = left + e.pageX - startX,
					y = top  + e.pageY - startY

				x = x.constrain(0, dw - w)
				y = y.constrain(0, dh - h)

				if (xDragable) {
					follow.css('left',  x)
				}
				
				if (yDragable) {
					follow.css('top',  y)
				}
			}

			document.on('mousemove', onmove, this)
			document.once('mouseup', function (e) {
				document.off('mousemove', onmove, this)
			}, this)
		}, this)
	},


	stop: function () {
		if (this.__anim) {
			var anim = this.__anim,
				cls = anim.class,
				handler = anim.handler,
				scope = anim.scope

			if (cls) {
				this.removeClass(cls)
			}
			
			if (handler) {
				this.off('transitionend', handler, scope)
			}
		}
	},


	__anim: null,


	//once the new animation enter, remove the previous anim
	//currently we don't need an animation sequence
	anim: function (className, isEnter, fn, scope) {
		var hasAnim = !! this.__anim
		if (hasAnim) {
			this.stop()
		}

		if (typeof isEnter === 'string') {
			isEnter = isEnter === 'enter'
		}

		if (isEnter) {
			//if already running animation, not reset start state
			if (!hasAnim) {
				this.css('transition', 'none')
				this.addClass(className + '-enter')
				this.redraw()
				this.css('transition', '')
				this.removeClass(className + '-enter')
			}

			var handle = function () {
				this.__anim = null
				if (fn) {
					fn.call(scope)
				}
			}

			this.__anim = {
				handler: handle,
				scope: this
			}

			this.once('transitionend', handle, this)
		}
		else {
			this.addClass(className + '-leave')

			var handle = function () {
				this.removeClass(className + '-leave')
				this.__anim = null
				if (fn) {
					fn.call(scope)
				}
			}

			this.__anim = {
				class: className + '-leave',
				handler: handle,
				scope: this
			}

			this.once('transitionend', handle, this)
		}
	},


	redraw: function () {
		this.offsetWidth
	},


	show: function () {
		this.hidden = false
		this.removeAttribute('hidden')
	},


	hide: function () {
		this.hidden = true
		this.setAttribute('hidden', 'hidden')
	},


	disable: function () {
		this.disabled = true
		this.setAttribute('disabled', 'disabled')
	},


	enable: function () {
		this.disabled = false
		this.removeAttribute('disabled')
	},


	visible: function () {
		return this.offsetWidth > 0 && this.offsetHeight > 0
	},
	

	destroy: function () {
		if (this.parentNode) {
			this.parentNode.removeChild(this)

			var node = this,
				below = true,
				nextNode

			node.offAll()

			do {
				if (below) {
					nextNode = node.firstElementChild || node.nextElementSibling
				}
				else {
					nextNode = node.nextElementSibling
				}

				if (nextNode) {
					nextNode.offAll()
					below = true
				}
				else {
					nextNode = node.parentNode
					below = false
				}

				node = nextNode
			}
			while (node && node !== this)
		}
	}
})



ff.applyIf(Element.prototype, {
	remove: function () {
		if (this.parentNode) {
			this.parentNode.removeChild(this)
		}
	}
})



var isMouseWheel       = 'onmousewheel' in window,
	isWebkitTransition = 'onwebkittransitionend' in window && !('ontransitionend' in window),
	isWebkitAnimation  = 'onwebkitanimationend'  in window && !('onanimationend'  in window)

var helper = ff.eventHelper = {

	EVENT_PREFIXS: {
		mousewheel: isMouseWheel ? 'mousewheel' : 'DOMMouseScroll',

		transition: isWebkitTransition ? 'WebkitTransition': 'transition',

		transitionend: isWebkitTransition ? 'webkitTransitionEnd': 'transitionend',

		animation: isWebkitAnimation ? 'WebkitAnimation': 'animation',

		animationend: isWebkitAnimation ? 'webkitAnimationEnd': 'animationend',

		animationiteration: isWebkitAnimation ? 'webkitAnimationIteration': 'animationiteration',
	},


	CUSTOM_EVENTS: {
		check: {
			change: function (e) {
				return e.target.checked === true
			}
		},

		uncheck: {
			change: function (e) {
				return e.target.checked === false
			}
		},

		mousewheelup: {
			mousewheel: function (e) {
				return e.detail < 0 || e.wheelDelta > 0
			}
		},

		mousewheeldown: {
			mousewheel: function (e) {
				return e.detail > 0 || e.wheelDelta < 0
			}
		},

		keydown: {
			keydown: function (e, t) {
				return !t || e.keyCode === helper.KEY_CODES[t]
			}
		},

		keyup: {
			keyup: function (e, t) {
				return !t || e.keyCode === helper.KEY_CODES[t]
			}
		},

		keypress: {
			keypress: function (e, t) {
				return !t || e.keyCode === helper.KEY_CODES[t]
			}
		}
	},


	KEY_CODES: {
		enter	: 13,
		tab		: 9,
		delete	: 46,
		up		: 38,
		left	: 37,
		right	: 39,
		down	: 40,
		esc		: 27
	},


	bindOnce: function (el, wrappedHandler, name, within, handler, scope, capture) {
		return function (e) {
			el.off(name, within, handler, scope, capture)
			wrappedHandler(e)
		}
	},


	bindWithin: function (el, wrappedHandler, within) {
		return function (e) {
			if (e.within(within)) {
				wrappedHandler(e)
			}
		}
	},


	bindWillCall: function (el, wrappedHandler, willCall, target) {
		return function (e) {
			if (willCall(e, target)) {
				wrappedHandler(e)
			}
		}
	},


	bindEvent: function (el, isOnce, name, within, handler, scope, capture) {
		var gestureType = helper.getGestureType(name)

		if (gestureType) {
			helper.bindGestureEvent(el, isOnce, gestureType, name, within, handler, scope)
			return
		}

		if (helper.isIntervalEvent(el, name)) {
			helper.bindIntervalEvent(el, isOnce, name, handler, scope, capture)
			return
		}

		var nativeName = name,
			wrappedHandler = handler,
			nativeEvents = {},
			target

		if (scope) {
			wrappedHandler = handler.bind(scope)
		}

		if (name.contains(':')) {
			target = name.after(':')
			nativeName = name = name.before(':')
		}

		if (isOnce) {
			wrappedHandler = helper.bindOnce(el, wrappedHandler, name, within, handler, scope, capture)
		}

		if (within) {
			wrappedHandler = helper.bindWithin(el, wrappedHandler, within)
		}

		if (name in helper.CUSTOM_EVENTS) {
			nativeEvents = {}

			for (var nativeName in helper.CUSTOM_EVENTS[name]) {
				var willCall = helper.CUSTOM_EVENTS[name][nativeName]

				nativeName = helper.EVENT_PREFIXS[nativeName] || nativeName
				nativeEvents[nativeName] = helper.bindWillCall(el, wrappedHandler, willCall, target)
			}
		}
		else {
			nativeName = helper.EVENT_PREFIXS[nativeName] || nativeName
			nativeEvents[nativeName] = wrappedHandler
		}


		//property with the $ prefix is additional property for Element
		el.__events = el.__events || {}
		el.__events[name] = el.__events[name] || []

		el.__events[name].push({
			el: el,
			name: name,
			nativeEvents: nativeEvents,
			within: within,
			target: target,
			handler: handler,
			scope: scope,
			capture: capture
		})

		for (var nativeName in nativeEvents) {
			el.addEventListener(nativeName, nativeEvents[nativeName], capture)
		}
	},


	GESTURES: ['tap', 'pan', 'pinch', 'press', 'rotate', 'swipe'],


	getGestureType: function (name) {
		if (window.Hammer) {
			for (var i in helper.GESTURES) {
				if (name.startsWith(helper.GESTURES[i])) {
					return helper.GESTURES[i]
				}
			}
		}

		return ''
	},


	bindGestureEvent: function (el, isOnce, gestureType, name, within, handler, scope) {
		var mc = el.__hammer
		if (!mc) {
			mc = el.__hammer = new Hammer.Manager(el === document ? el.documentElement : el)
		}

		var recognizer = mc.get(gestureType)

		if (!recognizer) {
			recognizer = new Hammer[gestureType.capitalize()]()
			recognizer.recognizeWith(mc.recognizers)
			mc.add(recognizer)
		}

		var wrappedHandler = handler
		if (scope) {
			wrappedHandler = handler.bind(scope)
		}

		if (isOnce) {
			wrappedHandler = helper.bindOnce(el, wrappedHandler, name, within, handler, scope, false)
		}

		if (within) {
			wrappedHandler = helper.bindWithin(el, wrappedHandler, within)
		}

		wrappedHandler = (function (e) {
			e.within = UIEvent.prototype.within
		}).seq(wrappedHandler)

		var nativeEvents = {}
		nativeEvents[name] = wrappedHandler

		el.__events = el.__events || {}
		el.__events[name] = el.__events[name] || []

		el.__events[name].push({
			el: el,
			name: name,
			nativeEvents: nativeEvents,
			within: '',
			target: '',
			handler: handler,
			scope: scope,
			capture: false,
			isGesture: true
		})

		mc.on(name, wrappedHandler)
	},


	intervalTime: 50,

	intervalTasks: {},

	interval: null,

	intervalSeed: 1,

	INTERVAL_EVENTS: ['show', 'hide', 'toggle', 'resize', 'boxchange'],


	isIntervalEvent: function (el, name) {
		return helper.INTERVAL_EVENTS.contains(name) && el !== window
	},


	bindIntervalEvent: function (el, isOnce, name, handler, scope, triggerAtOnce) {
		var e = {
			el: el,
			name: name,
			within: '',
			target: '',
			handler: handler,
			scope: scope,
			capture: false,
			isOnce: isOnce,
			lastState: undefined
		}

		var changed = this.checkIntervalState(e, triggerAtOnce)
		if (!changed || !e.isOnce) {
			el.__events = el.__events || {}
			el.__events[name] = el.__events[name] || []
			el.__events[name].push(e)

			var id = el.__intervalTaskId
			if (!id) {
				id = el.__intervalTaskId = helper.intervalSeed++
			}

			helper.intervalTasks[id] = helper.intervalTasks[id] || []
			helper.intervalTasks[id].push(e)

			helper.interval = helper.interval || new ff.Interval(helper.intervalTime, helper.runIntervalTasks)
			if (!helper.interval.timing) {
				helper.interval.reset()
			}
		}
	},


	//show, hide may trigger at once, if forceChange, will trigger event at once
	checkIntervalState: function (e, forceChange) {
		var el = e.el,
			name = e.name,
			handler = e.handler,
			scope = e.scope,
			lastState = e.lastState,
			newState,
			changed = false

		switch(name) {
			case 'show':
				newState = el.visible()
				changed = newState && newState != lastState
				break

			case 'hide':
				newState = !el.visible()
				changed = newState && newState != lastState
				break

			case 'toggle':
				newState = el.visible()
				changed = lastState !== undefined && newState !== lastState
				break

			case 'resize':
				newState = {
					width  : el.offsetWidth,
					height : el.offsetHeight,
				}

				changed = lastState !== undefined && !ff.deeplyEqual(newState, lastState)
				break

			case 'boxchange':
				newState = el.box(true)
				changed = lastState !== undefined && !ff.deeplyEqual(newState, lastState)
				break
		}

		e.lastState = newState
		changed = changed || forceChange

		if (changed) {
			handler.call(scope, {
				target: el,
				type: name,
				state: newState,
			})
		}

		return changed
	},


	runIntervalTasks: function () {
		var total = 0

		for (var i in helper.intervalTasks) {
			var tasks = helper.intervalTasks[i]

			for (var j = 0; j < tasks.length; j++) {
				var e = tasks[j],
					changed = helper.checkIntervalState(e)

				if (changed && e.isOnce) {
					tasks.splice(j--, 1)
				}
			}

			if (tasks.length === 0) {
				delete helper.intervalTasks[i]
			}

			total += tasks.length
		}

		if (total === 0) {
			helper.interval.close()
		}
	},


	/*will apply to element and window, document*/
	//scope is optional, it's only available when be object type, include null
	//last argument capture equals runAtOnce when running interval tasks
	on: ff.flex('string [string] function [object] [boolean]', function (name, within, handler, scope, capture) {
		var names = name.split(/\s+/)
		for (var i = 0; i < names.length; i++) {
			helper.bindEvent(this, false, names[i], within, handler, scope, capture)
		}

		return this
	}),


	once: ff.flex('string [string] function [object] [boolean]', function (name, within, handler, scope, capture) {
		var names = name.split(/\s+/)
		for (var i = 0; i < names.length; i++) {
			helper.bindEvent(this, true, names[i], within, handler, scope, capture)
		}

		return this
	}),


	//if scope is not exist, unbind all the handler, no matter what the scope is
	//if no target, unbind all the keyevent with any target
	//if no function, unbind all events by event name
	//if no within...
	off: ff.flex('string [string] [function] [object] [boolean]', function (name, within, handler, scope, capture) {
		if (this.__events && this.__events[name]) {
			var es = this.__events[name],
				target,
				isIntervalEvent = helper.isIntervalEvent(this, name),
				intervalTaskId = this.__intervalTaskId

			if (name.contains(':')) {
				target = name.after(':')
				name = name.before(':')
			}

			if (handler === ff.self) {
				handler = null
			}

			for (var i = es.length - 1; i >= 0; i--) {
				var e = es[i]

				if ((!handler || e.handler === handler) && (!within || e.within === within) && (!scope || e.scope === scope) && (!target || e.target === target) && e.capture === capture) {
					if (e.isGesture) {
						if (this.__hammer) {
							for (var nativeName in e.nativeEvents) {
								this.__hammer.off(nativeName, e.nativeEvents[nativeName])
							}
						}
					}
					else if (isIntervalEvent) {
						var tasks = helper.intervalTasks[intervalTaskId]
						if (tasks) {
							tasks.remove(e)
						}
					}
					else {
						for (var nativeName in e.nativeEvents) {
							this.removeEventListener(nativeName, e.nativeEvents[nativeName], capture)
						}
					}

					es.splice(i, 1)
				}
			}
		}
		else {
			this.removeEventListener(name, handler, capture)
		}

		return this
	}),


	offAll: function () {
		if (this.__events) {
			var intervalTaskId = this.__intervalTaskId

			if (intervalTaskId) {
				delete helper.intervalTasks[intervalTaskId]
			}

			for (var name in this.__events){
				var es = this.__events[name],
					isIntervalEvent = helper.isIntervalEvent(this, name)
	
				for (var i = es.length - 1; i >= 0; i--) {
					var e = es[i]

					if (e.isGesture) {
						if (this.__hammer) {
							for (var nativeName in e.nativeEvents) {
								this.__hammer.off(nativeName, e.nativeEvents[nativeName])
							}
						}
					}
					else if (!isIntervalEvent) {
						for (var nativeName in e.nativeEvents) {
							this.removeEventListener(nativeName, e.nativeEvents[nativeName], e.capture)
						}
					}
				}
			}
		}

		return this
	},


	hasEvent: ff.flex('string function [object] [boolean]', function (name, handler, scope, capture) {
		if (this.__events && this.__events[name]) {
			var es = this.__events[name],
				target

			if (name.contains(':')) {
				target = name.after(':')
				name = name.before(':')
			}

			if (!handler) {
				return es && es.length > 0
			}

			if (es && handler) {
				for (var i = es.length - 1; i >= 0; i--) {
					var e = es[i]

					if ((!handler || e.handler === handler) && (!scope || e.scope === scope) && (!target || e.target === target) && e.capture === capture) {
						return true
					}
				}
			}
		}

		return false
	}),


	trigger: function (name, e) {
		var gestureType = helper.getGestureType(name),
			isIntervalEvent = helper.isIntervalEvent(this, name)

		if (gestureType) {
			if (window.Hammer) {
				var e = e || {
						target: this
					},
					node = this

				while (node) {
					if (node.__hammer) {
						node.__hammer.emit(name, e)
					}
					node = node.parentNode
				}
			}
		}
		else if (isIntervalEvent) {
			var es = this.__events ? this.__events[name] : null

			if (es) {
				for (var i = es.length - 1; i >= 0; i--) {
					var e = es[i],
						changed = helper.checkIntervalState(e, true)

					if (changed && e.isOnce) {
						es.splice(i--, 1)
					}
				}
			}
		}
		else {
			var target = '',
				nativeName = name

			if (name.contains(':')) {
				target = name.after(':')
				name = nativeName = name.before(':')
			}
			else if(name in helper.CUSTOM_EVENTS) {
				nativeName = ff.keys(helper.CUSTOM_EVENTS[name])[0]
			}

			nativeName = helper.EVENT_PREFIXS[nativeName] || nativeName

			var e = document.createEvent('HTMLEvents')
			e.initEvent(nativeName, true, true)

			if (target in helper.KEY_CODES) {
				e.keyCode = helper.KEY_CODES[target]
			}
			else if (name === 'mousewheelup') {
				e.detail = -3
				e.wheelDelta = 120
			}
			else if (name === 'mousewheeldown') {
				e.detail = 3
				e.wheelDelta = -120
			}

			this.dispatchEvent(e)
		}

		return this
	},


	registerTo: function (Class) {
		ff.define(Class.prototype, helper, [
			'on',
			'once',
			'off',
			'offAll',
			'hasEvent',
			'trigger'
		])
	}
}

helper.registerTo(Window)
helper.registerTo(Document)
helper.registerTo(DocumentFragment)
helper.registerTo(Element)



var elementProperties = [
	'select',
	'selectAll',
	'selectValue',
	'query',
	'queryAll',
	'queryValue'
]
ff.define(Document.prototype, Element.prototype, elementProperties)
ff.define(DocumentFragment.prototype, Element.prototype, elementProperties)



ff.define(CharacterData.prototype, Element.prototype, [
	'before',
	'after',
	'parent',
	'destroy'
])

ff.define(CharacterData.prototype, {
	'get index': Object.getOwnPropertyDescriptor(Element.prototype, 'index').get
})

ff.applyIf(CharacterData.prototype, Element.prototype, [
	'remove'
])



ff.define(UIEvent.prototype, {

	//support node, or string argument for is
	within: function (el, until) {
		if (el instanceof Node) {
			return el.contains(this.target)
		}
		else {
			return !!this.target.up(el, until)
		}
	}
})

if (window.TouchEvent) {
	ff.define(TouchEvent.prototype, UIEvent.prototype, ['within'])
}

if (window.PointerEvent) {
	ff.define(PointerEvent.prototype, UIEvent.prototype, ['within'])
}



/*ff.ajax({
	method: POST or GET or others
	url: make sure has been trimed
	params: paramters of ajax request
	data: post data
	form: form data, will be send as urlencoded type, otherwise it will replace data property as data to send
	onload: handler when load successfully, arguments are response body and res
	onfail: handler when got error, arguments are response body and res
	onend: load or fail, arguments are response body and res
	scope: scope for handlers
	type: text, json, xml, html, buffer, blob
	timeout: how many seconds before timeout
	headers: the request headers
	[DEPRECATED] form: string or object, auto set method to post
	accept: text, json or html
)

events: load, fail, end. first argument is response body, second is res object:
	status: contains xhr.status and additional error text
	body: response body
	ok: is success
	finalURL: dest url which build from url and params
	config: initial config
*/
ff.Ajax = ff.extend(ff.Emitter, {

	TYPES: {
		text: 'text',
		json: 'text',
		xml: 'document',
		html: 'text',
		buffer: 'arraybuffer',
		blob: 'blob'
	},

	ACCEPT_TYPES: {
		text: 'text/plain',
		json: 'application/json',
		html: 'text/html'
	},

	url: '',

	method: 'GET',

	type: 'text',

	data: null,

	accept: 'text',

	params: null,

	timeout: 0,

	status: 0,

	//boolean or a hash
	cache: false,

	//request headers
	headers: null,

	//will cache config info
	config: null,

	//to capture upload progress, will force use options request when cross domain
	isUpload: false,

	init: function (c) {
		ff.apply(this, this.config = c)
		ff.Emitter.call(this)

		if (this.form || this.data) {
			this.method = 'POST'
		}

		this.type = this.type in this.TYPES ? this.type : 'text'

		if (this.onload) {
			this.on('load', this.onload, this.scope)
		}

		if (this.onfail) {
			this.on('fail', this.onfail, this.scope)
		}

		if (this.onend) {
			this.on('end', this.onend, this.scope)
		}
	},


	initXHR: function () {
		this.xhr = new XMLHttpRequest()
		this.xhr.addEventListener('load', this.onLoad.bind(this), false)
		this.xhr.addEventListener('error', this.onFail.bind(this), false)
		this.xhr.addEventListener('timeout', this.onTimeout.bind(this), false)
		this.xhr.addEventListener('progress', this.onProgress.bind(this), false)
		
		if (this.isUpload) {
			this.xhr.upload.addEventListener('progress', this.onUploadProgress.bind(this), false)
		}
	},


	getFinalURL: function () {
		var isQueried = /\?/.test(this.url),
			url = encodeURI(this.url)

		for (var i in this.params) {
			url += (isQueried ? '&' : '?') + i + '=' + encodeURIComponent(this.params[i])
			isQueried = true
		}

		return url
	},


	'get finalURL': function () {
		return this.getFinalURL()
	},


	getPostData: function () {
		if (this.data instanceof Array || String(this.data) === '[object Object]') {
			//cause cross domain issue because server may not set Access-Control-Allow-Headers: Content-Type
			//this.xhr.setRequestHeader('content-type', 'application/json')
			return JSON.stringify(this.data)
		}
		else {
			return this.data
		}
	},


	request: function () {
		this.initXHR()
		this.xhr.open(this.method, this.finalURL, true)
		this.xhr.responseType = this.TYPES[this.type]
		this.xhr.setRequestHeader('accept', this.ACCEPT_TYPES[this.accept] || this.ACCEPT_TYPES.text)

		for (var i in this.headers) {
			this.xhr.setRequestHeader(name, value)
		}

		if (this.timeout) {
			this.xhr.timeout = this.timeout
		}

		this.xhr.send(this.getPostData())

		return this
	},


	onLoad: function (e) {
		var res = this.getBaseRes()

		if (this.xhr.status >= 200 && this.xhr.status < 300) {
			res.ok = true

			var resType = (this.xhr.getResponseHeader('Content-Type') || '').trim().toLowerCase()

			if (this.type === 'json' || resType.endsWith('/json')) {
				try {
					res.body = res.body ? JSON.parse(res.body) : null
				}
				catch (e) {
					res.status = 'INVALID JSON'
					res.ok = false
				}
			}
		}

		this.onEnd(res)
	},


	getBaseRes: function (status) {
		var res = {
			ok: false,
			finalURL: this.finalURL,
			status: status || this.xhr.status,
			body: this.xhr.response,
			config: this.config
		}

		return res
	},


	onEnd: function (res) {
		this.trigger(res.ok ? 'load' : 'fail', res.body, res)
		this.trigger('end', res.body, res)
	},


	onFail: function (e) {
		var res = this.getBaseRes()
		this.onEnd(res)
	},


	//timeout not trigger error event
	onTimeout: function (e) {
		var res = this.getBaseRes('TIMEOUT')
		this.onEnd(res)
	},


	onProgress: function (e) {
		this.trigger('progress', e.loaded, e.total)
	},


	onUploadProgress: function (e) {
		this.trigger('uploadprogress', e.loaded, e.total)
	},


	abort: function () {
		if (this.xhr.readyState > this.xhr.UNSENT) {
			this.xhr.abort()
			var res = this.getBaseRes('ABORT')
			this.trigger('abort', res)
		}
	}
})



ff.apply(ff, {

	ajax: function (c) {
		return new ff.Ajax(c).request()
	},


	ajaxGet: function (url, onend, scope) {
		return ff.ajax({
			url: url,
			onend: onend,
			scope: scope
		})
	},


	ajaxPost: function (url, data, onend, scope) {
		return ff.ajax({
			url: url,
			method: 'POST',
			data: data,
			onend: onend,
			scope: scope
		})
	},


	//if you wang to capture upload progress, use this
	upload: function (url, data, onend, scope) {
		return ff.ajax({
			url: url,
			method: 'POST',
			isUpload: true,
			data: data,
			onend: onend,
			scope: scope
		})
	},


	download: function (urls) {
		if (typeof urls === 'string') {
			urls = [urls]
		}

		if (!urls.length) {
			return
		}

		var isChrome = navigator.userAgent.contains('Chrome') && !ff.IE
		
		if (isChrome) {
			while (urls.length) {
				var url = urls.shift()
				var a = document.create('<a href="{0}" download="{0}" hidden></a>'.format(url))
				document.body.appendChild(a)
				a.click()
				a.remove()
			}
		}
		else {
			ff.interval(200, function () {
				var url = urls.shift()
				var iframe = document.create('<iframe src="{0}" hidden></iframe>'.format(url))
				document.body.appendChild(iframe)
				iframe.onload = function () {
					iframe.destroy.defer(1000, iframe)
				}

				if (!urls.length) {
					return false
				}
			})
		}
	},


	postDownload: function (url, post) {
		var form = document.create('<form target="_blank" method="post" action="{0}"></form>'.format(url))

		for (var name in post) {
			var value = post[name],
				input = document.create('<input type="hidden">')

			input.name = name
			input.value = value

			form.append(input)
		}

		document.body.append(form)
		form.submit()
		form.destroy()
	},


	downloadText: function (text, type, fileName) {
		var blob = new Blob([text], {type: type})
		var fs = new FileReader

		if (ff.IE > 0) {
			var win = window.open('', "_blank", "")
			win.document.open()
			win.document.write(text)
			win.document.execCommand("saveAs", false, fileName)
			win.close()
		}
		else {
			fs.onload = function (e) {
				var a = document.create('<a download="{0}"></a>'.format(fileName))
				a.href = fs.result
				document.body.append(a)
				a.click()

				ff.timeout(1000, function () {
					a.destroy()
				})
			}

			fs.readAsDataURL(blob)
		}
	},


	//type support image, video, audio or others
	selectFile: ff.flex('[string] [boolean]', function (type, multiple, fn, scope) {
		var isChrome = navigator.userAgent.contains('Chrome') && !ff.IE
		if (isChrome) {
			type = '*'
		}

		var input = document.create('<input type="file" accept="{0}" {1} hidden>'.format(
			type || '*',
			multiple ? 'multiple' : ''
		))

		input.on('change', function (e) {
			fn.call(scope, e.target.files)
		})

		document.once('focus', function () {
			input.destroy()
		})
		
		document.body.append(input)
		input.click()
	}),


	selectFolder: ff.flex('[string]', function (type, fn, scope) {
		if (type === 'video/*') {
			type += ',video/mp4'
		}

		var input = document.create('<input type="file" accept="{0}" hidden webkitdirectory directory>'.format(
			type || '*'
		))

		input.on('change', function (e) {
			fn.call(scope, e.target.files)
		})

		document.once('focus', function () {
			input.destroy()
		})
		
		document.body.append(input)
		input.click()
	}),


	//fn got preload successfully, and img as arguments
	//defalt wait for 10s
	//retryTimes will reload when overtime
	defaultPreloadTimeout: 30000,


	preload: ff.flex('string [number] [number] [function]', function (url, timeout, retryTimes, fn, scope) {
		var img = new Image

		var handler = function (ok, e) {
			img.destroy()

			if (timerHandler.overtime && retryTimes > 0) {
				retryTimes -= 1
				ff.preload(url, timeout, retryTimes, fn, scope)
			}
			else {
				fn.call(scope, ok ? img : null)
			}
		}
		
		var timerHandler = handler.wait(timeout || ff.defaultPreloadTimeout)
		var loadHandler  = timerHandler.bind(null, true)
		var errorHandler = timerHandler.bind(null, false)

		img.on('load', loadHandler)
		img.on('error', errorHandler)
		img.src = url

		return function () {
			img.off('load', loadHandler)
			img.off('error', errorHandler)
			img.destroy()
			timerHandler.unWait()
		}
	}),


	//check localhost ip
	getLocalhostIp: function (callback) {

		try {
			var RTCPeerConnection = window.RTCPeerConnection || window.mozRTCPeerConnection || window.webkitRTCPeerConnection
			if (!RTCPeerConnection) {
				return callback('');
			}
			var pc = new RTCPeerConnection({iceServers:[]}), noop = function(){};      
            pc.createDataChannel(''); // create a bogus data channel
            pc.createOffer(pc.setLocalDescription.bind(pc), noop); // create offer and set local description
            pc.onicecandidate = function(ice) {
                if (ice && ice.candidate && ice.candidate.candidate) {
                    var myIP = /([0-9]{1,3}(\.[0-9]{1,3}){3}|[a-f0-9]{1,4}(:[a-f0-9]{1,4}){7})/.exec(ice.candidate.candidate)[1];
                    callback(myIP);
                    pc.onicecandidate = noop;
                }
            }.bind(this);
		} catch (e) {
			callback('');
		}
	},

	//when mouse leave all for 200ms, already known that mouse in first of els
	// onAllLeave: function (els, time, callback, scope) {
	// 	var mouseIn = true
	// 	var timeout

	// 	function enter () {
	// 		mouseIn = true
	// 	}

	// 	function leave () {
	// 		mouseIn = false

	// 		timeout = setTimeout(function () {
	// 			timeout = 0
	// 			if (!mouseIn) {
	// 				trigger()
	// 			}
	// 		}, time)
	// 	}

	// 	function trigger () {
	// 		callback.call(scope)
	// 	}

	// 	for (var i = 0; i < els.length; i++) {
	// 		var el = els[i]
	// 		el.on('mouseenter', enter)
	// 		el.on('mouseleave', leave)
	// 	}

	// 	return function () {
	// 		if (timeout) {
	// 			clearTimeout(timeout)
	// 		}

	// 		for (var i = 0; i < els.length; i++) {
	// 			var el = els[i]
	// 			el.off('mouseenter', enter)
	// 			el.off('mouseleave', leave)
	// 		}
	// 	}
	// },


	// //auto unbind after leave
	// onceAllLeave: function (els, time, callback, scope) {
	// 	var mouseIn = true
	// 	var timeout = 0

	// 	function enter () {
	// 		mouseIn = true
	// 	}

	// 	function leave () {
	// 		mouseIn = false

	// 		timeout = setTimeout(function () {
	// 			timeout = 0
	// 			if (!mouseIn) {
	// 				unbind()
	// 				trigger()
	// 			}
	// 		}, time)
	// 	}

	// 	function trigger () {
	// 		callback.call(scope)
	// 	}

	// 	function unbind () {
	// 		if (timeout) {
	// 			clearTimeout(timeout)
	// 		}

	// 		for (var i = 0; i < els.length; i++) {
	// 			var el = els[i]
	// 			el.off('mouseenter', enter)
	// 			el.off('mouseleave', leave)
	// 		}
	// 	}

	// 	for (var i = 0; i < els.length; i++) {
	// 		var el = els[i]
	// 		el.on('mouseenter', enter)
	// 		el.on('mouseleave', leave)
	// 	}

	// 	return unbind
	// }
})



ff.Resource = ff.extend({

	//url can contains {param} which come from action data (cant from default params config)
	//can use {.} to replace to string type of params
	url: '',

	//url as relative url to this base url
	base: '',

	//default
	method: 'GET',

	//default
	type: 'json',

	//default
	timeout: 0,

	//this params or params in action can contains {arg}
	params: null,

	//an object, can return data for testing for each action
	//when use interceptor, no ajax instance returns
	interceptor: null,

	//actions can contains params, data
	//when call action, args can use to format params and url
	//data argument can used to format or replace data
	actions: {
		'get': {
			method: 'GET'
		},
		'query': {
			method: 'GET'
		},
		'save': {
			method: 'POST'
		},
		'update': {
			method: 'PUT'
		},
		'delete': {
			method: 'DELETE'
		}
	},


	init: function (cfg) {
		ff.apply(this, cfg)

		for (var action in this.actions) {
			this[action] = this.newAction(action)
		}

		if (this.base && this.base.endsWith('/')) {
			this.base = this.base.slice(0, -1)
		}
	},


	//resolve from base url
	resolve: function (url) {
		url = url || ''

		if (/https?:\/\//.test(url) || !this.base) {
			return url
		}

		if (/^\//.test(url)) {
			var host = this.base.select(/^(\w+:\/*[^\/]*)/, 1)
			return host + url
		}
		else {
			return this.base + '/' + url
		}
	},


	newAction: function (action) {
		var method = this.actions[action].method || this.method || 'GET'

		if (method === 'GET') {
			return this.onGetAction.bind(this, action)
		}
		else {
			return this.onOtherAction.bind(this, action)
		}
	},


	getConfig: function (action, args, onend, scope) {
		var cfg = ff.clone(this.actions[action])
		ff.applyIf(cfg, this, ['url', 'method', 'type', 'timeout'])
		cfg.params = ff.clone(cfg.params)
		ff.applyIf(cfg.params, this.params)
		cfg.url = this.formatString(cfg.url, args)
		cfg.url = this.resolve(cfg.url)

		if (args) {
			for (var i in cfg.params) {
				cfg.params[i] = this.formatString(String(cfg.params[i]), args)
			}
		}

		cfg.onend = onend
		cfg.scope = scope

		return cfg
	},


	getFinalURL: function (action, paramData) {
		var cfg = this.getConfig(action, paramData)
		return ff.Ajax.prototype.getFinalURL.call(cfg)
	},


	formatString: function (s, json) {
		return s.replace(/\{([\w.]+)\}/g, function ($0, $1) {
			var v = $1 === '.' ? json : ff.own(json, $1)
			v = String(v)

			return v
		})
	},


	deeplyFormat: function (o, data) {
		if (typeof o === 'string' && /{\w+}/.test(o)) {
			if (/^{\w+}$/.test(o)) {
				o = data[o]
			}
			else if (o === '{.}') {
				o = data
			}
			else {
				o = this.formatString(o, data)
			}
		}

		ff.each(o, function (v, i) {
			if (typeof v === 'string' && /{\w+}/.test(v)) {
				if (/^{\w+}$/.test(v)) {
					o[i] = data[v]
				}
				else if (v === '{.}') {
					o[i] = data
				}
				else {
					o[i] = this.formatString(v, data)
				}
			}

			else if (typeof v === 'object') {
				this.deeplyFormat(v, data)
			}
		})

		return o
	},


	formatData: function (cfg, action, postData) {
		if (cfg.data && postData) {
			cfg.data = this.deeplyFormat(cfg.data, postData)
		}
		else {
			cfg.data = postData
		}
	},


	onGetAction: ff.flex('string [object|string|number]', function (action, args, onend, scope) {
		var cfg = this.getConfig(action, args, onend, scope)

		if (this.interceptor) {
			this.onInterceptor(action, cfg)
		}
		else {
			return ff.ajax(cfg)
		}
	}),


	//post data is not optional
	onOtherAction: ff.flex('string [object|string|number] object|string', function (action, args, postData, onend, scope) {
		var cfg = this.getConfig(action, args, onend, scope)
		this.formatData(cfg, action, postData)

		if (this.interceptor) {
			this.onInterceptor(action, cfg)
		}
		else {
			return ff.ajax(cfg)
		}
	}),


	onInterceptor: function (action, cfg) {
		var body = this.interceptor[action]

		if (body instanceof Function) {
			body = body(cfg)
		}

		var ok = body !== undefined,
			status = ok ? 200 : 404

		var res = {
			ok: ok,
			finalURL: ff.Ajax.prototype.getFinalURL.call(cfg),
			status: status,
			body: body,
			config: cfg
		}

		cfg.onend.bind(cfg.scope, body, res).defer()
	}
})



//onnext must returns ajax 
ff.Uploader = ff.extend(ff.Queue, {

	loaded: 0,

	total: 100,

	name: '',

	size: 0,

	init: function () {
		ff.Queue.apply(this, arguments)
	},


	onPrepareTask: function (file) {
		this.loaded = 0
		this.total = 100
		this.name = file.name
		this.size = file.size
	},


	onNextReturns: function (ajax) {
		ajax.on('uploadprogress', this.onUploadProgress, this)
	},


	onUploadProgress: function (loaded, total) {
		this.loaded = loaded
		this.total = total
	}
})



ff.store = {

	prefix: 'ff_',

	isSupported: function () {
		var key = this.prefix + 'support_test'

		try{
			localStorage[key] = 1
			delete localStorage[key]
			return true
		}
		catch (e) {
			return false
		}
	},

	has: function (key) {
		if (!this.isSupported()) {
			return false
		}

		key = this.prefix + key
		return key in localStorage
	},


	get: function (key) {
		if (!this.isSupported()) {
			return null
		}

		key = this.prefix + key

		var value = localStorage[key]
		if (value && typeof value === 'string') {
			try{
				value = JSON.parse(value)
				var expires = localStorage[key + '_expires']
				if (expires && expires < Date.now()) {
					delete localStorage[key]
					delete localStorage[key + '_expires']
					return null
				}
				else {
					return value
				}
			}
			catch (err) {
				return null
			}
		}
		else {
			return null
		}
	},


	set: function (key, value, expires) {
		if (!this.isSupported()) {
			return
		}

		key = this.prefix + key
		localStorage[key] = JSON.stringify(value)

		if (expires) {
			if (expires > 0) {
				localStorage[key + '_expires'] = Date.now() + expires * 1000
			}
		}
	},


	remove: function (key) {
		if (!this.isSupported()) {
			return
		}

		key = this.prefix + key

		delete localStorage[key]
		delete localStorage[key + '_expires']
	},


	clear: function () {
		if (!this.isSupported()) {
			return
		}

		if (localStorage) {
			localStorage.clear()
		}
	}
}

})();