/*
Extend base classes in javascript
*/

(function () {
'use strict'


/*Extend Object, ff is a short alias for Object*/
var ff = {

	//apply attributes from o to c, c can be a function, return o
	//if a, only apply properties in it
	//not work for hash
	apply: function (o, c, a) {
		if (a instanceof Array) {
			for (var i in a) {
				var p = a[i]
				o[p] = c[p]
			}
		}
		else {
			for (var p in c) {
				o[p] = c[p]
			}
		}

		return o
	},


	//apply only when o has no this property
	applyIf: function (o, c, a) {
		if (a instanceof Array) {
			for (var i in a) {
				var p = a[i]
				if (!(p in o) && (p in c)) {
					o[p] = c[p]
				}
			}
		}
		else {
			for (var p in c) {
				if (!(p in o)) {
					o[p] = c[p]
				}
			}
		}

		return o
	},


	clone: function (o, a) {
		return ff.apply({}, o, a)
	},


	deeplyClone: function (o) {
		if (typeof o !== 'object' || !o) {
			return o
		}

		var c
		if (o instanceof Array) {
			c = []
		}
		else {
			c = {}
		}

		for (var i in o) {
			c[i] = ff.deeplyClone(o[i])
		}

		return c
	},


	//only compare properties from a, deep default to 1
	deeplyEqual: function (a, b, deep) {
		if (typeof a !== 'object' || typeof b !== 'object' || !a || !b) {
			return a === b
		}

		for (var i in a) {
			if(typeof a[i] === 'object' && deep > 1) {
				return ff.deeplyEqual(a[i], b[i], deep)
			}

			else if (a[i] !== b[i]) {
				return false
			}
		}

		return true
	},


	//used to define properties
	defineValue: function (o, p, v, enumerable) {
		var definedO = {
			enumerable: !!enumerable,
			configurable: true,
			value: v,
			writable: true
		}

		Object.defineProperty(o, p, definedO)
	},


	//define a unenumerable getter
	defineGetter: function (o, p, v, enumerable) {
		var definedO = {
			enumerable: !!enumerable,
			configurable: true,
			get: v
		}

		Object.defineProperty(o, p, definedO)
	},


	//define unenumable values, the applied getter or setter will can't be found in own properties
	//not work for hash
	define: function (o, c, a) {
		if (a instanceof Array) {
			for (var i in a) {
				var p = a[i]
				ff.defineValue(o, p, c[p])
			}
		}
		else {
			for (var p in c) {
				if (p.slice(0, 4) === 'get ') {
					ff.defineGetter(o, p.slice(4), c[p])
				}
				else {
					ff.defineValue(o, p, c[p])
				}
			}
		}

		return o
	},


	defineIf: function (o, c, a) {
		if (a instanceof Array) {
			for (var i in a) {
				var p = a[i]
				if (!(p in o) && (p in c)) {
					ff.defineValue(o, p, c[p])
				}
			}
		}
		else {
			for (var p in c) {
				if (!(p in o)) {
					if (p.slice(0, 4) === 'get ') {
						ff.defineGetter(o, p.slice(4), c[p])
					}
					else {
						ff.defineValue(o, p, c[p])
					}
				}
			}
		}

		return o
	},


	//return an empty object to store data
	//since conflict with observer, suggest you not use it when doing observing on it
	hash: function () {
		return Object.create(null)
	},


	//create an object, which inherit proto, and overwritten by overwrite
	create: function (proto, overwrite) {
		var o = Object.create(proto)

		if (overwrite) {
			ff.apply(o, overwrite)
		}

		return o
	},


	//if has own property, return the related value from own properties
	//if is hash, return o[p]
	own: function (o, p) {
		if (o.hasOwnProperty) {
			if (o.hasOwnProperty(p)) {
				return o[p]
			}
			else {
				return undefined
			}
		}
		else {
			return o[p]
		}
	},


	hasOwn: function (o, p) {
		if (o.hasOwnProperty) {
			return o.hasOwnProperty(p)
		}
		else {
			return p in o
		}
	},


	//return an class which inherit c, and overwrite by o, c is optional
	//init will be the constructor, if not set, use super class init function
	extend: function (superClass, o) {
		if (!o) {
			o = superClass
			superClass = null
		}

		var Class = o.init,
			proto = superClass ? ff.create(superClass.prototype, o) : o

		for (var p in proto) {
			if (p.slice(0, 4) === 'get ') {
				ff.defineGetter(proto, p.slice(4), proto[p])
			}
		}

		if(superClass && !Class) {
			Class = proto.init = function () {
				superClass.apply(this, arguments)
			}
		}
		else if(!Class) {
			Class = proto.init = function () {}
		}

		Class.prototype = proto

		return Class
	},


	//return self, can be used as filter, also used as a empty function
	self: function (v) {
		return v
	},


	//get value by a string key or a function, not support scope
	get: function (item, key) {
		if (!item) {
			return undefined
		}
		else if (key instanceof Function) {
			return key(item)
		}
		else {
			return item[key]
		}
	},


	getAsync: function (item, key, callback) {
		if (!item) {
			return undefined
		}
		else if (key instanceof Function) {
			return key(item, callback)
		}
		else {
			return callback(item[key])
		}
	}
}



//event emit super class, no need to call ff.Emitter[.prototype.init].call(this), it will be called automaticially.
ff.Emitter = function () {
	//may be called more than once in one class, so need to check __events
	if (!this.__events) {
		this.__events = {}
		if (this.events) {
			for (var i in this.events) {
				this.on(i, this.events[i], this.events.scope || null)
			}
		}
	}
}


ff.Emitter.prototype = {

	init: ff.Emitter,

	events: null,

	on: function (name, handler, scope) {
		var e = this.__events[name] || (this.__events[name] = [])

		e.push({
			handler: handler,
			scope: scope || null
		})

		return this
	},


	once: function (name, handler, scope) {
		var e = this.__events[name] || (this.__events[name] = [])

		e.push({
			handler: handler,
			scope: scope || null,
			once: true
		})

		return this
	},


	//if scope is omit, unbind all the handler, no matter what the scope is
	off: function (name, handler, scope) {
		var e = this.__events[name]

		if (e) {
			for (var i = e.length - 1; i >= 0; i--) {
				if (e[i].handler === handler && (!scope || e[i].scope === scope)) {
					e.splice(i, 1)
				}
			}
		}

		return this
	},


	//if handler is omit, return true if event exist with this name
	hasListener: function (name, handler, scope) {
		var e = this.__events[name]

		if (!handler) {
			return e && e.length > 0
		}

		else if (e && handler) {
			for (var i = 0; i < e.length; i++) {
				if (e[i].handler === handler && (!scope || e[i].scope === scope)) {
					return true
				}
			}
		}

		return false
	},


	//return false if any handler return false, else return true
	trigger: function (name /* , arg1, arg2, ...*/ ) {
		var e = this.__events[name],
			returnValue = true

		if (e) {
			var args = ff.toArray(arguments).slice(1)

			for (var i in e) {
				var ei = e[i]

				//the handler may call off, so must remove it before handle
				if (ei.once === true) {
					e.splice(i--, 1)
				}
				if (ei.handler.apply(ei.scope, args) === false) {
					returnValue = false
				}
			}
		}

		return returnValue
	},


	destroy: function () {
		this.__events = {}
	}
}



/*Extend Argument*/
ff.apply(ff, {

	DEFAULT_ARGS: {
		string: '',
		number: 0,
		boolean: false,
		function: ff.self,
		object: null
	},

	__flexFormatsRECache: {},

	__initFlexFormats: function (formatsString) {
		if (ff.__flexFormatsRECache[formatsString]) {
			return ff.__flexFormatsRECache[formatsString]
		}

		var formats = formatsString.split(/\s+/),
			reString = '',
			defaults = []

		for (var i in formats) {
			var format = formats[i],
				optional = /\[.+?\]/.test(format)

			if (optional) {
				format = format.slice(1, -1)
			}

			var types = format.split('|'),
				shortFormat = types.map(function (t) {
					return t[0]
				}).join('|')

			reString += '(' + shortFormat + ')'

			if (optional) {
				reString += '?'
			}

			defaults[i] = ff.DEFAULT_ARGS[types[0]]
		}

		return ff.__flexFormatsRECache[formatsString] = {
			defaults: defaults,
			formatString: formatsString,
			re: new RegExp(reString)
		}
	},


	/*flexArgs(arguments, 'array [number] [function] [object]')
	support:
		string
		number
		boolean
		function
		object
	*/
	flexArgs: function (args, formats) {
		args = ff.toArray(args)

		if (typeof formats === 'string') {
			formats = ff.__initFlexFormats(formats)
		}

		var re = formats.re,
			defaults = formats.defaults

		var shortTypesString = args.map(function (arg) {
			return (typeof arg)[0]
		}).join('')

		var matches = shortTypesString.matches(re)
		if (matches.length === 0) {
			throw 'Flexable arguments `{0}` not match format `{1}`'.format(args.join(', '), formats.formatString)
		}

		for (var i = 0; i < matches.length; i++) {
			var match = matches[i]

			if (!match) {
				args.splice(i, 0, defaults[i])
			}
		}

		return args
	},


	flex: function (formats, fn, scope) {
		formats = ff.__initFlexFormats(formats)

		return function () {
			var args = ff.flexArgs(arguments, formats)

			return fn.apply(scope || this, args)
		}
	}
})



ff.define(Function.prototype, {

	//create a new function, which will call this function, then another function, return the returns of the last function
	//the last argument can be the scope, optional. if not set, take the scope when calling
	seq: function (fn) {
		var fns = [this],
			args = ff.toArray(arguments),
			scope = null

		if (typeof args[args.length - 1] === 'object') {
			scope = args.pop()
		}

		for (var i = 0; i < args.length; i++) {
			if (args[i] instanceof Function) {
				fns.push(args[i])
			}
		}

		return function () {
			var returns

			for (var i = 0; i < fns.length; i++) {
				returns = fns[i].apply(scope || this, arguments)
			}

			return returns
		}
	},


	//in milseconds
	//call clearTimeout to clear
	defer: function (ms, scope) {
		if (typeof ms !== 'number') {
			scope = ms
			ms = 0
		}
		else {
			ms = ms || 0
		}

		var fn = scope ? this.bind(scope) : this,
			df = setTimeout

		if (ms === 0 && typeof setImmediate !== 'undefined') {
			df = setImmediate
		}

		return df(fn, ms)
	},


	//return a new function, call it will call self
	//if have waited for ms milseconds, then call self with null as argument, calling returned fn will not call self any more
	//fn has an overtime property, and an original
	wait: function (ms, scope) {
		var self = this,
			timeoutId

		var fn = function () {
			if (!fn.overtime) {
				clearTimeout(timeoutId)
				timeoutId = 0
				self.apply(scope || this, arguments)
			}
		}

		timeoutId = setTimeout(function () {
			fn.overtime = true
			self.call(scope, null)
		}, ms)

		fn.overtime = false
		fn.canceled = false

		fn.unWait = function () {
			clearTimeout(timeoutId)
			fn.canceled = true
		}

		return fn
	},


	//return a new function, call it twice in a specify interval will only work for once
	//[DEPRECATED] restartTimerAfterCalled, call fn in blocking will restart timeout
	//if makesureLast is true, like scrolling, if same callback has been blocked, call it after end
	//returned fn has a timeoutId, blocking and original property
	block: ff.flex('number [boolean] [function]', function (ms, makesureLast, onblock, scope) {
		var self = this,
			lastArgs = null,
			lastScope = null,
			timeoutId = 0

		var timeoutFn = function () {
			timeoutId = 0
			fn.blocking = false

			if (makesureLast && lastArgs) {
				self.apply(scope || lastScope, lastArgs)
				lastArgs = null
				lastScope = null
			}
		}

		var fn = function () {
			if (fn.blocking) {
				lastArgs = arguments
				lastScope = this
				onblock.apply(scope || this, arguments)
			}
			else {
				timeoutId = setTimeout(timeoutFn, ms)
				fn.blocking = true
				lastArgs = null
				lastScope = null
				self.apply(scope || this, arguments)
			}
		}

		fn.blocking = false

		return fn
	}),


	debounce: function (ms, scope) {
		var self = this,
			timeoutId = 0

		var fn = function () {
			if (timeoutId) {
				clearTimeout(timeoutId)
			}

			var args = arguments,
				currentScope = scope || this 

			timeoutId = setTimeout(function (args) {
				timeoutId = 0
				self.apply(currentScope, args)
			}, ms)
		}

		return fn
	}
})



ff.defineIf(Array.prototype, {

	contains: function (e) {
		return this.indexOf(e) > -1
	}
})



ff.define(Array.prototype, {

	//get by index, support negative value
	at: function (i) {
		if (i < 0) {
			i += this.length
		}

		return this[i]
	},


	first: function () {
		return this[0]
	},


	last: function () {
		return this[this.length - 1]
	},


	//add unique value
	add: function () {
		for (var i = 0; i < arguments.length; i++) {
			var e = arguments[i],
				index = this.indexOf(e)

			if (index === -1) {
				this.push(e)
			}
		}

		return this
	},


	remove: function (e) {
		for (var i = 0; i < arguments.length; i++) {
			var e = arguments[i],
				index = this.indexOf(e)

			if (index > -1) {
				this.splice(index, 1)
			}
		}

		return this
	},


	empty: function () {
		this.splice(0, this.length)

		return this
	},


	pushAll: function (a) {
		this.push.apply(this, a)
		return this
	},


	unshiftAll: function (a) {
		this.unshift.apply(this, a)
		return this
	},


	addAll: function (a) {
		this.add.apply(this, a)
		return this
	},


	removeAll: function (a) {
		this.remove.apply(this, a)
		return this
	},


	//support [[]].join(\t, \n)
	joinBy: function (a, inner, outer) {
		if (typeof outer === 'string') {
			var out = []

			for (var i = 0; i < a.length; i++) {
				if (a[i] instanceof Array) {
					out.push(a[i].call(inner))
				}
				else {
					out.push(a[i])
				}
			}

			return out.call(outer)
		}
		else {
			return a.call(inner)
		}
	}
})



//for collections, both array and object, will overwrite
var collectionMethods = {

	find: function (o, fn, scope) {
		if (o instanceof Array) {
			for (var i = 0; i < o.length; i++) {
				if (fn.call(scope, o[i], i)) {
					return o[i]
				}
			}
		}
		else {
			for (var i in o) {
				if (fn.call(scope, o[i], i)) {
					return o[i]
				}
			}
		}

		return undefined
	},


	findAll: function (o, fn, scope) {
		return ff.filter(o, fn, scope)
	},


	isEmpty: function (o) {
		if (o instanceof Array) {
			return o.length
		}
		else {
			for (var i in o) {
				return false
			}
		}

		return true
	},


	empty: function (o) {
		if (o instanceof Array) {
			while (o.length) {
				o.pop()
			}
		}
		else {
			for (var i in o) {
				delete o[i]
			}
		}

		return o
	},


	//arguments or NodeList transform to array, or others to array
	toArray: function (o) {
		return Array.prototype.slice.call(o, 0)
	},


	//[[a], [b], c, null] -> [a, b, c]
	flatten: function (o) {
		var a = []

		ff.each(o, function(e) {
			if (e instanceof Array) {
				a.pushAll(e)
			}
			else if (e !== null && e !== undefined) {
				a.push(e)
			}
		})

		return a
	},


	//index for array is string type, for Array.prototype.each is number
	each: function (o, fn, scope) {
		if (o instanceof Array) {
			for (var i = 0; i < o.length; i++) {
				fn.call(scope, o[i], i)
			}
		}
		else {
			for (var i in o) {
				fn.call(scope, o[i], i)
			}
		}

		return o
	},


	keys: function (o) {
		var a = []

		ff.each(o, function(e, i) {
			a.push(i)
		})

		return a
	},


	values: function (o) {
		var a = []

		ff.each(o, function(e) {
			a.push(e)
		})

		return a
	},


	//{a:b} -> [[a, b]]
	pairs: function (o) {
		var a = []

		ff.each(o, function(e, i) {
			a.push([i, e])
		})

		return a
	},


	//{a:b} -> {b:a}
	invert: function (o) {
		var io = {}

		ff.each(o, function(e, i) {
			io[e] = i
		})

		return io
	},


	//the index argument is string, not number even for array, belows methods use this rule too
	every: function (o, fn, scope) {
		if (o instanceof Array) {
			for (var i = 0; i < o.length; i++) {
				if (!fn.call(scope, o[i], i)) {
					return false
				}
			}
		}
		else {
			for (var i in o) {
				if (!fn.call(scope, o[i], i)) {
					return false
				}
			}
		}

		return true
	},


	filter: function (o, fn, scope) {
		var a = []

		ff.each(o, function(e, i) {
			if (fn.call(scope, e, i)) {
				a.push(e)
			}
		})

		return a
	},


	filterObject: function (o, fn, scope) {
		var newO = {}

		ff.each(o, function(e, i) {
			if (fn.call(scope, e, i)) {
				newO[i] = e
			}
		})

		return newO
	},


	map: function (o, fn, scope) {
		var a = []

		ff.each(o, function(e, i) {
			a.push(fn.call(scope, e, i))
		})

		return a
	},


	mapObject: function (o, fn, scope) {
		var newO = {}

		ff.each(o, function(e, i) {
			newO[i] = fn.call(scope, e, i)
		})

		return newO
	},


	some: function (o, fn, scope) {
		if (o instanceof Array) {
			for (var i = 0; i < o.length; i++) {
				if (fn.call(scope, o[i], i)) {
					return true
				}
			}
		}
		else {
			for (var i in o) {
				if (fn.call(scope, o[i], i)) {
					return true
				}
			}
		}

		return false
	},


	// reduce: function (o,  fn, startValue, scope) {
	// 	var value = startValue

	// 	for (var i in o) {
	// 		value = fn.call(scope, value, o[i], i)
	// 	}

	// 	return value
	// },


	//([[a, b], [c, d]], 1) -> [b, d]
	//([[a, b], [c, d]], [0, 1]) -> [[a, b], [c, d]]
	pluck: function (o, col) {
		var a = []

		if (col instanceof Array) {
			ff.each(o, function(row) {
				if (row) {
					var pluckedRow = []

					for (var j = 0; j < col.length; j++) {
						pluckedRow.push(row[col[j]])
					}

					a.push(pluckedRow)
				}
			})
		}
		else {
			ff.each(o, function(row) {
				if (row) {
					a.push(row[col])
				}
			})
		}

		return a
	},


	//[[a,b], [c,d]] -> [a,c], [b,d]
	zip: function (a) {
		var zip = []

		for (var i = 0; i < a.length; i++) {
			zip[i] = []
		}

		for (var i = 0; i < arguments.length; i++) {
			var arg = arguments[i]
			for (var j = 0; j < arg.length; j++) {
				zip[j][i] = arg[j]
			}
		}

		return zip
	},


	//output value
	sortBy: ff.flex('object [function|string|number] [object] [boolean]', function (o, fn, scope, isReverse) {
		var sa = [],
			isFn = fn instanceof Function

		ff.each(o, function(v, i) {
			sa.push([v, isFn ? fn.call(scope, v, i) : v[fn]])
		})

		//if can't compare, then return -1 to not swap elements
		sa.sort(function (a, b) {
			return isReverse ? (a[1] < b[1] ? 1 : -1) : (a[1] > b[1] ? 1 : -1)
		})

		return sa.pluck(0)
	}),


	/*Collection Methods*/
	//only support string type, if not, will parse them to string
	//in at least one argument
	union: function () {
		var hash = {},
			as = ff.toArray(arguments),
			un = []

		for (var i = 0; i < as.length; i++) {
			var a = as[i]

			ff.each(a, function(v) {
				hash[v] = true
			})
		}

		return ff.keys(hash)
	},


	//only support string type, if not, will parse them to string
	//in all the arguments
	intersect: function () {
		var hash = {},
			as = arguments,
			it = []

		for (var i = 0; i < as.length; i++) {
			var a = as[i]

			ff.each(a, function(v) {
				hash[v] = (hash[v] || 0) + 1
			})
		}

		for (var i in hash) {
			if (hash[i] === as.length) {
				it.push(i)
			}
		}

		return it
	},


	//only support string type, if not, will parse them to string
	//in the first, not in others
	diff: function () {
		var hash = ff.indexByValue(arguments[0]),
			as = ff.toArray(arguments).slice(1),
			df = []

		for (var i in as) {
			var a = as[i]

			ff.each(a, function(v) {
				if (v in hash) {
					delete hash[v]
				}
			})
		}

		for (var i in hash) {
			if (hash[i] === as.length) {
				df.push(i)
			}
		}

		return ff.keys(hash)
	},


	//only support string type, if not, will parse them to string
	unique: function (o) {
		return ff.keys(ff.indexByValue(o))
	},


	max: function (o) {
		if (!(o.length >= 0)) {
			o = ff.values(o)
		}

		return Math.max.apply(null, o)
	},


	min: function (o) {
		if (!(o.length >= 0)) {
			o = ff.values(o)
		}

		return Math.min.apply(null, o)
	},


	count: function (o) {
		ff.keys(o).length
	},


	sum: function (o) {
		var sum = 0

		ff.each(o, function(v) {
			sum += v
		})

		return sum
	},


	avg: function (o) {
		var sum = 0,
			len = 0

		ff.each(o, function(v) {
			sum += v
			len++
		})

		return sum / len || 0
	},


	first: function (o) {
		if (o.length >= 0) {
			return o[0]
		}
		else {
			for (var i in o) {
				return o[i]
			}
		}
	},


	last: function (o) {
		if (o.length >= 0) {
			return o[o.length - 1]
		}
		else {
			var v
			for (var i in o) {
				v = o[i]
			}
			return v
		}
	},


	/*SQL like syntaxs, use not strict equals*/
	//{a:b, c:d}, {a:b} -> {a:b, c:d}
	//the second object can be an object
	//use ==, not === in find
	select: function (rs, index, value) {
		var like
		if (typeof index === 'object') {
			like = index
		}
		else {
			like = {}
			like[index] = value
		}

		for (var i in rs) {
			var isLike = true

			for (var j in like) {
				if (rs[i][j] != like[j]) {
					isLike = false
					break
				}
			}

			if (isLike) {
				return rs[i]
			}
		}

		return null
	},


	//{a:b, c:d}, {a:b} -> [{a:b, c:d}]
	selectAll: function (rs, index, value) {
		var like
		if (typeof index === 'object') {
			like = index
		}
		else {
			like = {}
			like[index] = value
		}

		var a = []

		ff.each(rs, function (r) {
			var isLike = true

			for (var j in like) {
				var v1 = r[j],
					v2 = like[j]

				if (v1 != v2) {
					isLike = false
					break
				}
			}

			if (isLike) {
				a.push(r)
			}
		})

		return a
	},


	//index default as key
	//value default as value
	indexValueBy: function (rs, ip, vp, scope) {
		var hash = {},
			isIFn = ip instanceof Function,
			isVFn = vp instanceof Function

		ff.each(rs, function (r, i) {
			var index,
				value

			if (isIFn) {
				index = ip.call(scope, r, i)
			}
			else if(typeof ip === 'string' || typeof ip === 'number') {
				index = r[ip]
			}
			else{
				index = i
			}

			if (isVFn) {
				value = vp.call(scope, r, i)
			}
			else if(typeof vp === 'string' || typeof vp === 'number') {
				value = r[vp]
			}
			else {
				value = r
			}

			if (index !== null && index !== undefined) {
				hash[index] = value
			}
		})

		return hash
	},


	//index become value property or fn(value, key)
	indexBy: function (s, p, scope) {
		return ff.indexValueBy(s, p, undefined, scope)
	},


	//value become value property or fn(value, key)
	valueBy: function (s, p, scope) {
		return ff.indexValueBy(s, undefined, p, scope)
	},


	//index become value, if use fixValue, value will be it
	indexByValue: function (o, fixValue) {
		var hash = {}

		ff.each(o, function (v) {
			var index = v,
				value = fixValue === undefined ? v : fixValue

			if (index !== null && index !== undefined) {
				hash[index] = value
			}
		})

		return hash
	},


	swapIndexValue: function (o) {
		var hash = {}

		ff.each(o, function (v, i) {
			var index = v,
				value = i

			if (index !== null && index !== undefined) {
				hash[index] = value
			}
		})

		return hash
	},


	//several values encoded to one, joined by ,
	__encodeIndexes: function (vs) {
		for (var i in vs) {
			vs[i] = String(vs[i]).replace(/[\\,]/g, '\\$&')
		}

		return vs.join(',')
	},


	__decodeIndexes: function (v) {
		var vs = []
		var lastIndex = 0

		v.replace(/\\\\|\\,|,/g, function ($0, index) {
			if ($0 === ',') {
				vs.push(v.slice(lastIndex, index))
				lastIndex = index + $0.length
			}
		})

		vs.push(v.slice(lastIndex))

		for (var i in vs) {
			vs[i] = vs[i].replace(/\\([\\,])/g, '$1')
		}

		return vs
	},


	//like indexBy, but will return {index : records}
	//records array has a unenumable property __target, specify the object which contains all the required values that used to group
	groupBy: function (rs, cols, scope) {
		var hash = {}

		cols = typeof cols === 'string' || cols instanceof Function ? [cols] : cols
		cols = cols instanceof Array && typeof cols[0] === 'string' ? cols.indexByValue() : cols

		ff.each(rs, function (r) {
			var target = {},
				targetValues = [],
				encodedIndexes

			for (var mp in cols) {
				var p = cols[mp],
					v = p instanceof Function ? p.call(scope, r) : r[p]

				target[mp] = v
				targetValues.push(v)
			}

			encodedIndexes = targetValues.length === 1 ? String(targetValues[0]) : ff.__encodeIndexes(targetValues)

			var group = hash[encodedIndexes]
			if (!group) {
				group = hash[encodedIndexes] = []
				ff.defineValue(group, '__target', target)
			}

			group.push(r)
		})

		return hash
	},


	pushGroup: function (ga, gb) {
		for (var i in gb) {
			if (ga[i]) {
				ga[i].pushAll(gb[i])
			}
			else {
				ga[i] = gb[i]
			}
		}
	},


	unshiftGroup: function (ga, gb) {
		for (var i in gb) {
			if (ga[i]) {
				ga[i].unshiftAll(gb[i])
			}
			else {
				ga[i] = gb[i]
			}
		}
	},


	removeGroup: function (ga, gb) {
		for (var i in gb) {
			if (ga[i]) {
				ga[i].removeAll(gb[i])

				if (!ga[i].length) {
					delete ga[i]
				}
			}
		}
	},


	//count by property or fn
	countBy: function (rs, cols, scope) {
		var hash = ff.groupBy(rs, cols, scope)

		for (var i in hash) {
			hash[i] = hash[i].length
		}

		return hash
	},


	//ps is array of properties, or {Column : fn, Column : Column}
	//agg is like {TotalUser : {User : function (allUsers){}}} or {User : function (allRecords){}} or {TotalUser : [User, func...]}, or only a function
	aggregate: function (rs, cols, agg, scope) {
		var hash = ff.groupBy(rs, cols),
			results = []

		if (agg instanceof Function && typeof cols === 'string') {
			var mp = cols,
				fn = agg
			agg = {}
			agg[mp] = [mp, fn]
		}
		else {
			for (var mp in agg) {
				var fn = agg[mp]

				if (fn instanceof Function) {
					agg[mp] = [mp, fn]
				}
				else if (String(fn) === '[object Object]') {
					agg[mp] = ff.pairs(fn)[0]
				}
			}
		}

		for (var i in hash) {
			var group = hash[i],
				target = group.__target

			for (var mp in agg) {
				var a = agg[mp],
					p = a[0],
					fn = a[1],
					values = group

				if (typeof p === 'string') {
					values = group.pluck(p)
				}

				target[mp] = fn(values)
			}

			results.push(target)
		}

		return results
	},


	orderBy: function (rs, cols, scope) {
		if (typeof cols === 'string') {
			cols = [cols]
		}

		var reverse = []
		for (var i in cols) {
			if (/ desc$/.test(cols[i])) {
				reverse[i] = true
				cols[i] = cols[i].replace(/ desc$/, '')
			}
			else {
				reverse[i] = false
			}
		}

		var sa = []

		ff.each(rs, function (r) {
			var a = [r]

			for (var j in cols) {
				var p = cols[j],
					v = p instanceof Function ? p(r) : r[p]

				a.push(v)
			}

			sa.push(a)
		})

		var colsLen = cols.length

		sa.sort(function (a, b) {
			for (var i = 1; i <= colsLen; i++) {
				var ai = a[i],
					bi = b[i],
					rv = reverse[i - 1]

				if (rv) {
					return ai < bi ? 1 : -1
				}
				else if (ai < bi) {
					return ai > bi ? 1 : -1
				}
			}

			return -1
		})

		return sa.pluck(0)
	},
}



ff.apply(ff, collectionMethods)



ff.apply(ff, {
	indexOf: function (o, lookup) {
		ff.each(o, function (e, i) {
			if (e === lookup) {
				return i
			}
		})

		return undefined
	},


	contains: function (o, lookup) {
		return ff.indexOf(o, lookup) !== undefined
	}
})



ff.each(collectionMethods, function (method, property) {
	ff.defineValue(Array.prototype, property, function () {
		var args = ff.toArray(arguments)
		args.unshift(this)

		return ff[property].apply(null, args)
	})
})



ff.apply(Number, {

	range: function (a, b, s) {
		s = Math.abs(s) || 1

		var r = []

		if (a >= b) {
			for (var i = a; i >= b; i -= s) {
				r.push(i)
			}
		}
		else {
			for (var i = a; i <= b; i += s) {
				r.push(i)
			}
		}

		return r
	},


	//returns number between 0 to n(not include n)
	random: function (n) {
		n = n || 1

		return Math.floor(Math.random() * n)
	},


	//return a string joined by several random number
	randomString: function (l) {
		var s = ''

		do {
			s += String(Math.random()).replace('0.', '')
			if (s.length > l) {
				s = s.slice(0, l)
			}
		} while (s.length >= l)

		return s
	}
})



ff.define(Number.prototype, {
	//like toFixed, can be 0, -1, 1, the useless 0 after . will be removed
	fix: function (decimal) {
		decimal = decimal || 0

		var rate = Math.pow(10, decimal)

		return Math.round(this * rate) / rate
	},


	fillTo: function (len, char) {
		return String(this).fillTo(len, char || '0')
	},


	//arg1 <= this <= arg2, arg3 to force strictly
	between: function (a, b, strict) {
		if (a > b) {
			var c = a
			a = b
			b = c
		}

		if (strict) {
			return this > a && this < b
		}
		else {
			return this >= a && this <= b
		}
	},


	//force this to in the range, or
	constrain: function (a, b) {
		var n = Number(this),
			c

		if (a > b) {
			c = a
			a = b
			b = c
		}

		if (n < a) {
			n = a
		}
		else if (n > b) {
			n = b
		}

		return n
	},


	//congruence
	mod: function (n) {
		return (this % n + n) % n
	}
})



ff.apply(String, {

	CHARS: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz',

	random: function (l) {
		var s = ''

		for (var i = 0; i < l; i++) {
			s += this.CHARS[Math.floor(this.CHARS.length * Math.random())]
		}

		return s
	},


	//\r -> \\r, quote is current js quote which wrap current js piece, like ' or "
	jsEncode: function (value, quote) {
		value = String(value).replace(/[\r\n\t\v]/g, {
			'\\': '\\\\',
			'\r': '\\r',
			'\n': '\\n',
			'\t': '\\t',
			'\v': '\\v'
		})

		if (quote) {
			value = value.replace(new RegExp(quote, 'g'), '\\' + quote)
		}

		return value
	},


	//\\r -> \r
	jsDecode: function (value) {
		return String(value).replace(/\\['"rntv]/g, {
			"\\'": "'",
			'\\"': '"',
			'\\r': '\r',
			'\\n': '\n',
			'\\t': '\t',
			'\\v': '\v'
		})
	},


	//< -> &lt
	htmlEncode: function (value) {
		return String(value).replace(/&|>|<|"|'/g, {
			'&': '&amp;',
			'>': '&gt;',
			'<': '&lt;',
			'"': '&quot;',
			"'": '&apos;'
		})
	},


	//&lt -> <
	htmlDecode: function (value) {
		return String(value).replace(/&(amp|gt|lt|quot|apos|nbsp|copy);/g, {
			amp: '&',
			gt: '>',
			lt: '<',
			quot: '"',
			apos: "'",
			nbsp: '\xa0',
			copy: '\xa9'
		})
	}
})



ff.define(String.prototype, {

	//replace $0 to args[0], $1 to args[1], args is the arguments when replacing
	__replace$iByArgs: function (args) {
		return this.replace(/\$([$&\d])/g, function ($0, $1) {
			if ($1 === '$') {
				return '$'
			}

			else if ($1 === '&') {
				return args[0]
			}

			else if (ff.hasOwn(args, $1)) {
				return args[$1]
			}

			else {
				return $0
			}
		})
	},

	//tar can be an object, will replace $1 or $0 according the mapping rules specified by tar, and can replace $0 like $&
	replaceBy: function (re, tar) {
		if (typeof tar === 'string' && tar.indexOf('$0') > -1) {
			return this.replace(re, function () {
				return tar.__replace$iByArgs(arguments)
			})
		}

		else if (typeof tar === 'object') {
			return this.replace(re, function ($0, $1) {
				if (typeof $1 === 'string') {
					return tar[$1] || ''
				}
				else {
					return tar[$0] || ''
				}
			})
		}

		else {
			return this.replace(re, tar)
		}
	},


	//'12345'.select(/\d/) -> ['1']
	//'12345'.select(/(\d)/) -> ['1', '1']
	//'12345'.select(/(\d)/g) -> [['1'],['2'],['3'],['4'],['5']]
	//'12345'.select(/(\d)/, '$1') -> '1'
	//'12345'.select(/(\d)/, 1) -> '1'
	//'12345'.select(/(\d)/g, '$1$1') -> ['11','22','33','44','55']
	//'12345'.select(/(\d)/g, ['a', 'b', 'c', 'd', 'e', 'f']) -> ['b', 'c', 'd', 'e', 'f']
	//support $0
	//when replace by an object, only get value from own properties
	select: function (re, tar) {
		//use match is more quickly than replace
		if (re.global && tar === 0) {
			var matches = this.match(re)
			if (matches) {
				delete matches.index
			}
			else {
				matches = []
			}

			return matches
		}

		var m

		if (!re.global) {
			m = this.match(re) || []
			delete m.index

			if (typeof tar === 'number') {
				m = m[tar] || ''
			}

			else if (typeof tar === 'string') {
				m = tar.__replace$iByArgs(m)
			}

			else if (tar instanceof Function) {
				m = tar.apply(null, m) || ''
			}

			else if (tar instanceof Object) {
				m = ff.own(tar, m[1]) || ''
			}
		}

		else {
			m = []

			this.replace(re, function () {
				var args = ff.toArray(arguments).slice(0, -2)

				for (var i = 0; i < args.length; i++) {
					args[i] = args[i] || ''
				}

				m.push(args)
			})

			if (typeof tar === 'number') {
				for (var i = 0; i < m.length; i++) {
					m[i] = m[i][tar]
				}
			}

			else if (typeof tar === 'string') {
				for (var i = 0; i < m.length; i++) {
					m[i] = tar.__replace$iByArgs(m[i])
				}
			}

			else if (tar instanceof Function) {
				for (var i = 0; i < m.length; i++) {
					m[i] = tar.apply(null, m[i]) || ''
				}
			}

			else if (tar instanceof Object) {
				for (var i = 0; i < m.length; i++) {
					m[i] = ff.own(tar, m[i][1]) || ''
				}
			}
		}

		return m
	},


	matches: function (re) {
		var a = this.select(re)

		if (re.global) {
			for (var i = 0; i < a.length; i++) {
				a[i].shift()
			}
		}
		else {
			a.shift()
		}

		return a
	},


	//'a{0}{1}d'.format('b','c') -> 'abcd'
	//'a{x}{y}d'.format({x:'b',y:'c'}) -> 'abcd'
	//if keepProperty is true, keep it if no such property, used for formatting more than once
	format: function (json, keepProperty) {
		var o = json instanceof Object ? json : ff.toArray(arguments)

		return this.replace(/\{(\w+)\}/g, function ($0, $1) {
			var v = ff.own(o, $1)
			return keepProperty && !v ? $0 : String(v)
		})
	},


	//''.splitBy(' ') returns [], not ['']
	//'a\tb\n...'.splitBy(\n, \t)
	splitBy: function (s, ss) {
		var a

		if (!this.length) {
			a = []
		}
		else {
			a = this.split(s)
		}

		if (ss) {
			for (var i = 0; i < a.length; i++) {
				a[i] = a[i].split(ss)
			}
		}

		return a
	},


	before: function (slash, greedy) {
		var index = this.indexOf(slash)

		if (index < 0) {
			return greedy ? String(this) : ''
		}
		else {
			return this.slice(0, index)
		}
	},


	after: function (slash, greedy) {
		var index = this.indexOf(slash)

		if (index < 0) {
			return greedy ? String(this) : ''
		}
		else {
			return this.slice(index + 1)
		}
	},


	lastBefore: function (slash, greedy) {
		var index = this.lastIndexOf(slash)

		if (index < 0) {
			return greedy ? String(this) : ''
		}
		else {
			return this.slice(0, index)
		}
	},


	lastAfter: function (slash, greedy) {
		var index = this.lastIndexOf(slash)

		if (index < 0) {
			return greedy ? String(this) : ''
		}
		else {
			return this.slice(index + 1)
		}
	},


	fillTo: function (len, char) {
		char = char || ' '

		var str = this,
			len = len - str.length

		for (var i = 0; i < len; i++) {
			str = char + str
		}

		return str
	},


	capitalize: function () {
		return this.slice(0, 1).toUpperCase() + this.slice(1)
	},


	camelize: function () {
		return this.replace(/-([a-z])/g, function ($0, $1) {
			return $1.toUpperCase()
		})
	},


	deCamelize: function () {
		return this.replace(/[A-Z]/g, function ($0) {
			return '-' + $0.toLowerCase()
		})
	},


	contains: function (sub) {
		return this.indexOf(sub) > -1
	},


	get: function (i) {
		if (i < 0) {
			i += this.length
		}

		return this[i]
	},


	first: function () {
		return this.slice(0, 1)
	},


	last: function () {
		return this.slice(-1)
	},


	startsWith: function (s) {
		return this.slice(0, s.length) === s
	},


	endsWith: function (s) {
		return this.slice(-s.length) === s
	}
})



ff.apply(Date, {

	DEFAULT_FORMATS: 'y/M/d h:m:s',

	GET_METHODS: {
		y: 'getFullYear',
		M: 'getMonth',
		d: 'getDate',
		h: 'getHours',
		m: 'getMinutes',
		s: 'getSeconds'
	},

	SET_METHODS: {
		y: 'setFullYear',
		M: 'setMonth',
		d: 'setDate',
		h: 'setHours',
		m: 'setMinutes',
		s: 'setSeconds'
	},

	parseDate: function (d) {
		d = d.select(/\d+/g, 0)

		for (var i = 0; i < 6; i++) {
			d[i] = Number(d[i]) || 0
		}

		d[1] -= 1

		return new Date(d[0], d[1], d[2], d[3], d[4], d[5])
	},


	isValid: function (y, m, d, h, i, s, ms) {
		h = h || 0
		i = i || 0
		s = s || 0
		ms = ms || 0

		var dt = new Date(y, m - 1, d, h, i, s, ms)

		return y == dt.getFullYear() &&
			m == dt.getMonth() + 1 &&
			d == dt.getDate() &&
			h == dt.getHours() &&
			i == dt.getMinutes() &&
			s == dt.getSeconds() &&
			ms == dt.getMilliseconds()
	}
})


ff.define(Date.prototype, {
	//return a new date
	add: function (what, count) {
		if (!Date.SET_METHODS[what]) {
			throw 'Date adding types must be y, M, d, h, m, s'
		}

		var c = this[Date.GET_METHODS[what]](),
			d = this.clone()

		d[Date.SET_METHODS[what]](c + count)

		return d
	},


	clone: function () {
		return new Date(this.getTime())
	},


	format: function (f) {
		return (f || Date.DEFAULT_FORMATS).replace(/[yMdhms]/g, (function ($0) {
			var n = this[Date.GET_METHODS[$0]]()

			if ($0 === 'M') {
				n += 1
			}

			return n.fillTo(2)
		}).bind(this))
	},


	isLeapYear: function () {
		var year = this.getFullYear()
		return year % 4 === 0 && (year % 100 !== 0 || year % 400 === 0 && year)
	},


	getDaysOfYear: function () {
		return this.isLeapYear() ? 366 : 365
	},


	getDaysOfMonth: function () {
		var d = this.clone()
		d.setDate(32)

		return 32 - d.getDate()
	}
})



/*
support events:
	next(task): call after call onnext
	end(err): call after complete

queue is reuseable, it's never ended except meeting error or been close

There is still a defeat, processing tasks can't be the same
*/
ff.Queue = ff.extend(ff.Emitter, {

	concurrency: 1,

	//paused will cause new task not start, but already processing tasks will still be processed
	paused: false,

	onnext: null,

	onend: null,

	scope: null,

	timeout: 0,

	init: function (options) {
		ff.Emitter.apply(this)
		ff.apply(this, options)
		this.tasks = this.tasks || []
		this.processedTasks = []
		this.processingTasks = []
		this.cancelCalls = []
		this.onend = this.onend || ff.self
	},


	'get left': function () {
		return this.tasks.length
	},


	'get processing': function () {
		return this.processingTasks.length
	},


	'get processed': function () {
		return this.processedTasks.length
	},


	//running and not running
	'get unprocessed': function () {
		return this.processingTasks.length + this.tasks.length
	},


	unshift: function (task) {
		this.tasks.unshift(task)
		this.process()
	},


	push: function (task) {
		this.tasks.push(task)
		this.process()
	},


	pushAll: function (tasks) {
		this.tasks.pushAll(tasks)
		this.process()
	},

	unshiftAll: function (tasks) {
		this.tasks.unshiftAll(tasks)
		this.process()
	},


	pause: function () {
		this.paused = true

		return this
	},


	resume: function () {
		this.paused = false
		this.process()

		return this
	},


	process: function () {
		while (!this.paused && this.processing < this.concurrency && this.tasks.length > 0) {
			var task = this.tasks.shift()
			this.prepareNext(task)
		}

		return this
	},


	prepareNext: function (task) {
		var next

		if (this.timeout > 0) {
			next = this.onProcessed.bind(this, task).wait(this.timeout)
		}
		else {
			next = this.onProcessed.bind(this, task)
		}

		next.__index = this.processed + this.processing
		this.processingTasks.push(task)

		this.onPrepareTask(task)

		var cancelCall = this.onnext.call(this.scope, task, next),
			willCacheCall = this.processingTasks.contains(task) && (cancelCall instanceof Function || ff.Ajax && cancelCall instanceof ff.Ajax)

		if (willCacheCall) {
			this.cancelCalls.push({
				task: task,
				cancel: cancelCall instanceof Function ? cancelCall : cancelCall.abort.bind(cancelCall),
				ajax: ff.Ajax && cancelCall instanceof ff.Ajax ? cancelCall : null
			})
			this.onNextReturns(cancelCall)
		}

		this.trigger('next', task)
	},


	onPrepareTask: function () {},


	onNextReturns: function () {},


	start: function () {
		if (this.unprocessed === 0) {
			this.onend.call(this.scope, null)
		}
		else {
			this.process()
		}

		return this
	},


	onProcessed: function (task, err) {
		this.processingTasks.remove(task)
		this.processedTasks.push(task)

		var cancelCall = this.cancelCalls.select('task', task)
		if (cancelCall) {
			this.cancelCalls.remove(cancelCall)
		}

		if (err instanceof Error) {
			this.onProcessEnd(err)
		}
		else if (this.unprocessed === 0) {
			this.onProcessEnd(null)
		}
		else {
			this.process.defer(this)
		}
	},


	//will restore processing tasks
	onProcessEnd: function (err) {
		while(this.cancelCalls.length > 0) {
			this.cancelCalls.pop().cancel()
		}
		this.tasks.unshiftAll(this.processingTasks)
		this.processingTasks = []
		this.onend.bind(this.scope, err).defer()
		this.trigger('end', err)
	},


	//clear processed and not processing, but keep processing tasks
	clear: function () {
		this.tasks = []
		this.processedTasks = []
	},


	//clear processedTasks if not processing 
	reset: function () {
		if (!this.unprocessed) {
			this.processedTasks = []
		}
	},


	//force to end the sequence, and cancel processing tasks
	end: function (clear) {
		this.onProcessEnd(null)
		if (clear) {
			this.clear()
		}
		return this
	}
})



ff.apply(ff, {
	//fn   : (arguments from prevFn, next)
	//next : ([err], arguments for nextFn)
	//cb   : (err, last return)
	waterfall: ff.flex('object [function] [object]', function (fns, cb, scope) {
		var prevArgs = [];

		return new ff.Queue({
			tasks: fns,

			onnext: function (fn, next) {
				next = (function (err) {
					prevArgs = ff.toArray(arguments)
				}).seq(next)

				prevArgs.push(next)

				fn.apply(scope, prevArgs)
			},

			onend: function (err) {
				var args = err ? [err] : prevArgs
				cb.apply(scope, args)
			}
		}).start()
	}),


	//fns  : [fn(next)]
	//next : ([err])
	//cb   : (err, returns of fns)
	//run fns in several currencies, results catch the first argument of each next
	series: ff.flex('object [number] [function] [object]', function (fns, con, cb, scope) {
		var rs = []

		return new ff.Queue({
			concurrency: con || 1,
			tasks: fns,

			onnext: function (fn, next) {
				wrappedNext = (function (r) {
					rs[next.__index] = r
				}).seq(next)

				fn.call(scope, wrappedNext)
			},

			onend: function (err) {
				cb.call(scope, err || rs)
			}
		}).start()
	}),


	//fn   : (element, next)
	//next : ([err])
	//cb   : (err)
	eachAsync: ff.flex('object [number] function [function] [object]', function (a, con, fn, cb, scope) {
		return new ff.Queue({
			concurrency: con || 1,
			tasks: a,

			onnext: function (e, next) {
				fn.call(scope, e, next)
			},

			onend: function (err) {
				cb.call(scope, err)
			}
		}).start()
	}),


	//fn   : (element, next)
	//next : ([err], mapped value)
	//cb   : ([err], all mapped values)
	//results will keep the tasks order
	mapAsync: ff.flex('object [number] function [function] [object]', function (a, con, fn, cb, scope) {
		var rs = [];

		return new ff.Queue({
			concurrency: con || 1,
			tasks: a,

			onnext: function (e, next) {
				var wrappedNext = (function (r) {
					rs[next.__index] = r
				}).seq(next)

				fn.call(scope, e, wrappedNext)
			},

			onend: function (err) {
				cb.call(scope, err || rs)
			}
		}).start()
	}),


	//like map, but will concat returns
	concatAsync: ff.flex('object [number] function [function] [object]', function (a, con, fn, cb, scope) {
		ff.mapAsync(a, con, fn, function (rs) {
			rs = rs.flatten ? rs.flatten() : rs
			cb.call(scope, rs)
		}, scope)
	}),


	//fn   : (element, next)
	//next : (true or false)
	//cb   : (all filtered values)
	filterAsync: ff.flex('object [number] function [function] [object]', function (a, con, fn, cb, scope) {
		var fa = [];

		return new ff.Queue({
			concurrency: con || 1,
			tasks: a,

			onnext: function (e, next) {
				next = (function (r) {
					if (r) {
						fa.push(e)
					}
				}).seq(next)

				fn.call(scope, e, next)
			},

			onend: function () {
				cb.call(scope, fa)
			}
		}).start()
	}),


	//fn   : (element, next)
	//next : (true or false)
	//cb   : (is someone returns true)
	someAsync: ff.flex('object [number] function [function] [object]', function (a, fn, cb, scope) {
		var b = false

		var q = new ff.Queue({
			concurrency: con || 1,
			tasks: a,

			onnext: function (e, next) {
				next = (function (r) {
					if (r) {
						b = true
						q.end()
					}
				}).seq(next)

				fn.call(scope, e, next)
			},

			onend: function () {
				cb.call(scope, b)
			}
		}).start()

		return q
	}),


	//fn   : (element, next)
	//next : (true or false)
	//cb   : (is all elements returns true)
	everyAsync: ff.flex('object [number] function [function] [object]', function (a, con, fn, cb, scope) {
		var b = true

		var q = new ff.Queue({
			concurrency: con || 1,
			tasks: a,

			onnext: function (e, next) {
				next = (function (r) {
					if (!r) {
						b = false
						q.end()
					}
				}).seq(next)

				fn.call(scope, e, next)
			},

			onend: function () {
				cb.call(scope, b)
			}
		}).start()

		return q
	}),


	//fn   : (element, next)
	//next : (true or false)
	//cb   : (the finded value)
	findAsync: ff.flex('object [number] function [function] [object]', function (a, con, fn, cb, scope) {
		var v

		var q = new ff.Queue({
			concurrency: con || 1,
			tasks: a,

			onnext: function (e, next) {
				next = (function (r) {
					if (r) {
						v = e
						q.end()
					}
				}).seq(next)

				fn.call(scope, e, next)
			},

			onend: function () {
				cb.call(scope, v)
			}
		}).start()

		return q
	}),


	findAllAsync: function () {
		return ff.filterAsync.apply(this, arguments)
	},


	//fn return false to stop interval, please make sure you call next fn not immediatelly
	//fn got two arguments: next fn and current calling plus
	//call the returned fn to clear
	while: function (fn, scope) {
		var plus = 0,
			ended = false

		var next = function (nextArg) {
			if (nextArg !== false && !ended) {
				fn.call(scope, next, plus++)
			}
		}

		fn.call(scope, next, plus++)

		return function () {
			ended = true
		}
	},


	//return true to stop
	until: function (fn, scope) {
		var plus = 0,
			ended = false

		var next = function (nextArg) {
			if (nextArg !== true && !ended) {
				fn.call(scope, next, plus++)
			}
		}

		fn.call(scope, next, plus++)

		return function () {
			ended = true
		}
	},


	//in milseconds, fn return false to stop interval
	//call clearInterval to clear
	interval: function (ms, fn, scope) {
		var intervalId
		var intervalFn = function () {
			if (fn.call(scope) === false) {
				clearInterval(intervalId)
			}
		}

		return intervalId = setInterval(intervalFn, ms)
	},


	//a quick method for anonymous functions
	timeout: ff.flex('[number] function', function (ms, fn, scope) {
		return fn.defer(ms, scope)
	})
})



//arguments: ms, fn, scope
ff.Timer = ff.extend({

	ms: 0,

	fn: null,

	scope: null,

	/*inner properties*/
	id: 0,

	paused: false,

	'get timing': function () {
		return this.id > 0
	},


	init: ff.flex('[number] [boolean] function', function (ms, startImmediately, fn, scope) {
		this.fn = fn || ff.self
		this.ms = ms || 0
		this.scope = scope

		if (startImmediately) {
			this.reset()
		}
	}),


	reset: function () {
		if (this.paused) {
			return
		}

		this.close()

		if (this.ms >= 0 && this.ms < Infinity) {
			this.id = this.onTimeout.defer(this.ms, this)
		}
	},


	pause: function () {
		this.paused = true
		this.close()
	},


	resume: function () {
		this.paused = false
		this.reset()
	},


	onTimeout: function () {
		this.close()
		this.fn.call(this.scope)
	},


	setMs: function (ms) {
		this.ms = ms
		this.reset()
	},


	close: function () {
		if (this.id) {
			clearTimeout(this.id)
		}
		this.id = 0
	}
})



//arguments: ms, fn, scope
ff.Interval = ff.extend(ff.Timer, {

	reset: function () {
		this.close()

		var fn = this.fn
		if (this.scope) {
			fn = fn.bind(this.scope)
		}

		if (this.ms > 0 && this.ms < Infinity) {
			this.id = setInterval(fn, this.ms)
		}
	},


	close: function () {
		if (this.id) {
			clearInterval(this.id)
		}
		this.id = 0
	}
})



//used to cache callbacks with the same id
ff.CallbackGroup = ff.extend({

	data: null,

	init: function () {
		this.data = {}
	},


	has: function (id) {
		return !!this.data[id]
	},


	push: function (id, cb) {
		this.data[id] = this.data[id] || []
		this.data[id].push(cb)
	},


	call: function (id) {
		var args = ff.toArray(arguments).slice(1),
			cbs = this.data[id]

		for (var i = 0; i < cbs.length; i++) {
			var cb = cbs[i]
			cb.apply(this, args)
		}

		delete this.data[id]
	}
})



if (typeof window === 'object' && typeof document === 'object') {
	window.ff = ff
}

else {
	module.exports = ff
}


})();