(function () {
'use strict'


function mapAsync(tasks, onnext, onend) {
	var returns = []

	function nextTask() {
		if (tasks.length === 0) {
			onend(returns)
			return
		}

		onnext(tasks.shift(), function (r) {
			returns.push(r)
			nextTask()
		})
	}

	nextTask()
}



//define a module, if noNeedInit
window.define = function (name, module, inited) {
	var cache = require.cache
	cache[name] = module

	if (typeof module === 'function') {
		module.uninited = !inited
	}

	return module
}



//require module name, or
window.require = function (name, handler, scope) {
	var cache = require.cache

	if (typeof name === 'string') {
		var definedModule = cache[name],
			outModule = definedModule

		if (definedModule && definedModule.uninited) {
			var inModule = {exports: {}}
			definedModule(inModule, inModule.exports)
			outModule = cache[name] = inModule.exports
		}

		if (!outModule) {
			throw 'Can\'t find module ' + name
		}

		return outModule
	}

	else {
		var paths = name,
			names = []

		for (var i in paths) {
			names.push(i)
		}

		mapAsync(names, function (name, nextName) {
			var filePaths = paths[name]

			if (name in cache) {
				nextName(require(name))
			}

			else if (!filePaths) {
				throw 'Can\'t find module ' + name
			}

			else {
				if (typeof filePaths === 'string') {
					filePaths = [filePaths]
				}

				mapAsync(filePaths, function (path, nextPath) {
					var ext = path.select(/\.(\w+)$/, 1).toLowerCase(),
						tag

					var onerror = function () {
						nextPath(new Error('Error occured when loading ' + path))
					}

					if (ext === 'css') {
						tag = document.createElement('link')
						tag.rel = 'stylesheet'
						tag.href = path
					}
					else if (ext === 'js') {
						tag = document.createElement('script')
						tag.type = 'text/javascript'
						tag.src = path
					}

					if (ext === 'css' || ext === 'js') {
						tag.on('load', nextPath)
						tag.on('error', onerror)
						document.head.append(tag)
					}
					else {
						nextPath(new Error('module path ' + path + ' must ends with .js or .css'))
					}
				},

				function(){
					nextName(require(name))
				})
			}
		},

		function (modules) {
			handler.apply(scope, modules)
		})
	}
}


require.cache = {}

})();