ff.apply(ff, {

	XML2JSON: function (node, arrayItemName) {
		var json = {}

		if (!node) {
			return json
		}

		this._XMLNodeToJSON(node, json, arrayItemName)

		return json
	},


	//firefox will parse child text nodes which split with \n to multiple nodes
	_isAllChildNodesTextType: function (node) {
		var childNodes = node.childNodes

		if (!childNodes.length) {
			return false
		}

		for (var i = 0; i < childNodes.length; i++) {
			var child = childNodes[i]
			if (child.nodeType !== 3) {
				return false
			}
		}

		return true
	},


	_XMLNodeToJSON: function (node, json, arrayItemName) {
		if (node.nodeType === 9) {
			node = node.documentElement
		}

		if (node.nodeType === 1) {
			var name = node.nodeName,
				value

			if (node.attributes.length === 0 && ff._isAllChildNodesTextType(node)) {
				value = node.textContent
			}
			else if (node.childNodes.length === 0) {
				value = ''
			}
			else {
				value = {}

				for (var i = 0; i < node.attributes.length; i++) {
					value[node.attributes[i].nodeName] = node.attributes[i].nodeValue
				}

				for (var i = 0; i < node.childNodes.length; i++) {
					ff._XMLNodeToJSON(node.childNodes[i], value, arrayItemName)
				}
			}

			if (value instanceof Object && value[arrayItemName] && Object.keys(value).length === 1) {
				json[name] = value[arrayItemName]
			}
			else if (json[name]) {
				if (json[name] instanceof Array) {
					json[name].push(value)
				}
				else {
					json[name] = [json[name], value]
				}
			}
			else if (name === arrayItemName) {
				json[name] = [value]
			}
			else {
				json[name] = value
			}
		}

		//text nodes with element siblings
		else if (node.nodeType === 3) {
			var text = node.nodeValue
			if (text && /\S/.test(text)) {
				json["#text"] = text
			}
		}
	},


	//json must be an object, any can't include [[]] like
	JSON2XML: function (json, attrFirst, arrayItemName) {
		if (typeof attrFirst === 'string') {
			arrayItemName = attrFirst
			attrFirst = false
		}

		var dom = document.implementation.createDocument("", "", null)

		ff._JSONToXMLNode(json, dom, attrFirst, arrayItemName)

		return dom
	},


	_JSONToXMLNode: function (json, node, attrFirst, arrayItemName) {
		var dom = node.ownerDocument || node

		for (var name in json) {
			var value = json[name]

			if (typeof value === 'string' || typeof value === 'number' || typeof value === 'boolean') {
				if (attrFirst) {
					node.setAttribute(name, value)
				}
				else {
					var child = dom.createElement(name)
					child.appendChild(dom.createTextNode(String(value)))
					node.appendChild(child)
				}
			}

			else if (value instanceof Array) {
				if (arrayItemName) {
					var outer = dom.createElement(name)

					for (var i = 0; i < value.length; i++) {
						var item = value[i],
							child = dom.createElement(arrayItemName)

						if (typeof item === 'string' || typeof item === 'number' || typeof item === 'boolean') {
							child.appendChild(dom.createTextNode(String(item)))
						}
						else {
							ff._JSONToXMLNode(item, child, attrFirst, arrayItemName)
						}

						outer.appendChild(child)
					}

					node.appendChild(outer)
				}
				else {
					for (var i = 0; i < value.length; i++) {
						var item = value[i],
							child = dom.createElement(name)

						if (typeof item === 'string' || typeof item === 'number' || typeof item === 'boolean') {
							child.appendChild(dom.createTextNode(String(item)))
						}
						else {
							ff._JSONToXMLNode(item, child, attrFirst, arrayItemName)
						}

						node.appendChild(child)
					}
				}
			}

			else {
				var child = dom.createElement(name)
				ff._JSONToXMLNode(value, child, attrFirst, arrayItemName)

				node.appendChild(child)
			}
		}
	},


	//will ignore all empty nodes, ingnore
	XMLSerialize: function (node) {
		var output = []

		if (node.nodeType === 9) {
			node = node.documentElement
		}
		ff._XMLSerializeChildNode(node, output, '')

		return output.join('')
	},


	_encodeXMLString: function (text) {
		return text.replace(/&/g, '&amp;')
			.replace(/</g, '&lt;')
			.replace(/>/g, '&gt;')
			.replace(/"/g, '&quot;')
	},

	
	_XMLSerializeChildNode: function (node, output, tabs) {
		if (node.nodeType === 1) {
			output.push(tabs + '<' + node.tagName)

			for (var i = 0; i < node.attributes.length; i++) {
				var attr = node.attributes[i]
				output.push(' ' + attr.nodeName + '="' + ff._encodeXMLString(attr.nodeValue) + '"')
			}

			output.push('>')

			if (node.childNodes.length === 0) {
				output.push('</' + node.tagName + '>')
				output.push('\n')
			}
			else if (node.childNodes.length === 1 && node.firstChild.nodeType === 3) {
				output.push(ff._encodeXMLString(node.firstChild.nodeValue) + '</' + node.tagName + '>')
				output.push('\n')
			}
			else {
				output.push('\n')
				var nextTabs = tabs + '\t'
				for (var i = 0; i < node.childNodes.length; i++) {
					ff._XMLSerializeChildNode(node.childNodes[i], output, nextTabs)
				}
				output.push(tabs + '</' + node.tagName + '>')
				output.push('\n')
			}
		}
	}
});