(function () {
	'use strict'

	var __webpack_require__ = Vue._require
	var _ = __webpack_require__(11)
	var transition = __webpack_require__(50)



	_.apply = function (o, c) {
		for (var p in c) {
			o[p] = c[p]
		}
	}



	_.apply(_, {
		keys: function (o) {
			var a = []

			if (o instanceof Array) {
				for (var i = 0; i < o.length; i++) {
					a.push(i)
				}
			}
			else {
				for (var i in o) {
					a.push(i)
				}
			}

			return a
		},


		deeplyClone: function (o) {
			if (typeof o !== 'object' || !o) {
				return o
			}

			var c

			if (o instanceof Array) {
				c = []
				for (var i = 0; i < o.length; i++) {
					c[i] = _.deeplyClone(o[i])
				}
			}
			else {
				c = {}
				for (var i in o) {
					c[i] = _.deeplyClone(o[i])
				}
			}

			return c
		},


		block: function (fn, ms) {
			var lastArgs = null,
				lastScope = null,
				timeoutId = 0

			var timeoutFn = function () {
				timeoutId = 0

				if (lastArgs) {
					fn.apply(lastScope, lastArgs)
					lastArgs = null
					lastScope = null
				}
			}

			return function () {
				if (timeoutId) {
					lastArgs = arguments
					lastScope = this
				}
				else {
					timeoutId = setTimeout(timeoutFn, ms)
					lastArgs = null
					lastScope = null
					fn.apply(this, arguments)
				}
			}
		}
	})



	_.apply(Vue.config, {
		prefix : 'f-'
	})



	_.apply(Vue, {

		skip: function (o) {
			_.define(o, '_isVue', true)
		},


		//force to flush all the queued triggers
		flush: _.flush,


		//force to notify value changed in a callback, even though it's value is not changed
		notify: function (o) {
			if (!o) {
				return
			}

			var ob = o.__ob__

			ob.notify(true)
		}
	})



	_.apply(Vue.prototype, {

		$trigger: Vue.prototype.$emit,


		$filter: function (id) {
			return this.$options.filters[id]
		},


		$addComponent: function (cfg, com) {
			if (typeof cfg === 'string') {
				cfg = {
					el: cfg
				}
			}

			cfg._asComponent = true

			if (!com && cfg.el) {
				var el = cfg.el,
					tag

				if (typeof el === 'string') {
					el = document.query(el)
				}

				if (el) {
					tag = el.localName

					if (tag in Vue.options.components) {
						com = Vue.options.components[tag]
					}

					if (!com && el.getAttribute(Vue.config.prefix + 'component')) {
						tag = el.getAttribute(Vue.config.prefix + 'component')
						if (tag in Vue.options.components) {
							com = Vue.options.components[tag]
						}
						el.removeAttribute(Vue.config.prefix + 'component')
					}
				}
			}

			var child = this.$addChild(cfg, com),
				refID = child._refID

			if (refID) {
				this.$[refID] = child
			}

			return child
		}
	})



	//make class and style can be merged
	_.apply(_, {
		copyAttributes: function (from, to) {
			if (from.hasAttributes()) {
				var attrs = from.attributes
				for (var i = 0, l = attrs.length; i < l; i++) {
					var attr = attrs[i]

					if (to.hasAttribute(attr.name) && (attr.name === 'class' || attr.name === 'style')) {
						var newValue = to.getAttribute(attr.name)
						if (attr.name === 'style') {
							newValue += /;\s*$/.test(newValue) ? '' : '; '
						}
						else if (attr.name === 'class') {
							newValue += ' '
						}
						newValue += attr.value
						to.setAttribute(attr.name, newValue)
					}
					else {
						to.setAttribute(attr.name, attr.value)
					}
				}
			}
		},


		remove: function (el) {
			el.destroy()
		}
	})



	_.apply(Vue.options.directives, {

		show: function (value) {
			var el = this.el
			transition.apply(el, value ? 1 : -1, function () {
				el.hidden = !value
				if (value) {
					el.removeAttribute('hidden')
				}
				else {
					el.setAttribute('hidden', 'hidden')
				}
			}, this.vm)
		},


		hide: function (value) {
			var el = this.el
			transition.apply(el, value ? -1 : 1, function () {
				el.hidden = !!value
				if (!value) {
					el.removeAttribute('hidden')
				}
				else {
					el.setAttribute('hidden', 'hidden')
				}
			}, this.vm)
		},


		enable: function (value) {
			var el = this.el
			if (value) {
				el.enable()
			}
			else {
				el.disable()
			}
		},


		disable: function (value) {
			var el = this.el
			if (value) {
				el.disable()
			}
			else {
				el.enable()
			}
		},


		value: function (value) {
			var el = this.el
			el.value = value
		},


		//will set src a little time later, to avoid still show previous src before the next src load
		imgsrc: function (value, oldValue) {
			var el = this.el,
				self = this

			if (value && oldValue) {
				el.setAttribute('src', '')

				if (this.lazySrcSettingTimer) {
					clearTimeout(this.lazySrcSettingTimer)
				}

				this.lazySrcSettingTimer = setTimeout(function () {
					self.lazySrcSettingTimer = 0
					el.setAttribute('src', value || '')
				})
			}
			else {
				el.setAttribute('src', value || '')
			}
		}
	})



	var Watcher = __webpack_require__(25)
	var expParser = __webpack_require__(22)
	var literalRE = /^(true|false|\s?('[^']*'|"[^"]")\s?)$/

	Vue.directive('in', {

		priority: 900,

		bind: function () {
			var child = this.vm
			var parent = child.$parent
			var childKey = this.arg || '$data'
			var parentKey = this.expression

			if (this.el && this.el !== child.$el) {
				_.warn(
					'v-in can only be used on instance root elements.'
				)
			}
			else if (!parent) {
				_.warn(
					'v-in must be used on an instance with a parent.'
				)
			}
			else {
				var locked = false
				var lock = function () {
					locked = true
					_.nextTick(unlock)
				}
				var unlock = function () {
					locked = false
				}

				this.parentWatcher = new Watcher(
					parent,
					parentKey,
					function (val) {
						if (!locked) {
							lock()
							child.$set(childKey, val)
						}
					}
				)

				child.$set(childKey, this.parentWatcher.value)
			}
		},

		unbind: function () {
			if (this.parentWatcher) {
				this.parentWatcher.teardown()
			}
		}
	})



	Vue.directive('out', {

		priority: 900,

		bind: function () {
			var child = this.vm
			var parent = child.$parent
			var childKey = this.arg || this.expression
			var parentKey = this.expression

			if (this.el && this.el !== child.$el) {
				_.warn(
					'v-out can only be used on instance root elements.'
				)
			}
			else if (!parent) {
				_.warn(
					'v-out must be used on an instance with a parent.'
				)
			}
			else if (literalRE.test(parentKey)) {
				if (!this.arg) {
					_.warn(
						'v-in cannot bind literal value as $data: ' +
						parentKey
					)
				}
				else {
					var value = expParser.parse(parentKey).get()
					child.$set(childKey, value)
				}
			}
			else {
				var locked = false
				var lock = function () {
					locked = true
					_.nextTick(unlock)
				}
				var unlock = function () {
					locked = false
				}

				this.childWatcher = new Watcher(
					child,
					childKey,
					function (val) {
						if (!locked) {
							lock()
							parent.$set(parentKey, val)
						}
					}
				)

				parent.$set(parentKey, this.childWatcher.value)
			}
		},

		unbind: function () {
			if (this.childWatcher) {
				this.childWatcher.teardown()
			}
		}
	})




})(Vue);



(function () {

	var vueTouch = {}
	var Hammer = window.Hammer
	var gestures = ['tap', 'pan', 'pinch', 'press', 'rotate', 'swipe']
	var customeEvents = {}

	vueTouch.install = function (Vue) {

		Vue.directive('touch', {

			isFn: true,

			acceptStatement: true,

			bind: function () {
				if (!this.el.hammer) {
					this.el.hammer = new Hammer.Manager(this.el)
				}

				var mc = this.mc = this.el.hammer
					// determine event type
				var event = this.arg
				var recognizerType, recognizer

				if (customeEvents[event]) { // custom event
					var custom = customeEvents[event]
					recognizerType = custom.type
					recognizer = new Hammer[capitalize(recognizerType)](custom)
					recognizer.recognizeWith(mc.recognizers)
					mc.add(recognizer)
				}
				else {
					// built-in event
					for (var i = 0; i < gestures.length; i++) {
						if (event.indexOf(gestures[i]) === 0) {
							recognizerType = gestures[i]
							break
						}
					}

					if (!recognizerType) {
						console.warn('Invalid v-touch event: ' + event)
						return
					}

					recognizer = mc.get(recognizerType)

					if (!recognizer) {
						// add recognizer
						recognizer = new Hammer[capitalize(recognizerType)]()
							// make sure multiple recognizers work together...
						recognizer.recognizeWith(mc.recognizers)
						mc.add(recognizer)
					}
				}
			},

			update: function (fn) {
				var mc = this.mc
				var vm = this.vm
				var event = this.arg

				// teardown old handler
				if (this.handler) {
					mc.off(event, this.handler)
				}

				// define new handler
				this.handler = function (e) {
					e.targetVM = vm
					fn.call(vm, e)
				}

				mc.on(event, this.handler)
			},

			unbind: function () {
				this.mc.off(this.arg, this.handler)
				if (!Object.keys(this.mc.handlers).length) {
					this.mc.destroy()
					this.el.hammer = null
				}
			}
		})
	}

	/**
	 * Register a custom event.
	 *
	 * @param {String} event
	 * @param {Object} options - a Hammer.js recognizer option object.
	 *                           required fields:
	 *                           - type: the base recognizer to use for this event
	 */

	vueTouch.registerCustomEvent = function (event, options) {
		options.event = event
		customeEvents[event] = options
	}

	function capitalize(str) {
		return str.charAt(0).toUpperCase() + str.slice(1)
	}

	if (window.Vue && Hammer) {
		window.VueTouch = vueTouch
		Vue.use(vueTouch)
	}

})();